"""Test NetworkPoller GRPC infrastructure"""
import asyncio
import os
import sys
import unittest

from preseem_protobuf.network_poller.network_poller_pb2 import NetworkElementUpdate
from preseem import FakeNetworkPollerGrpcClient

class TestNetworkPollerGrpc(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.new_event_loop() # needed to initialize asyncio
        asyncio.set_event_loop(self.loop)
        self.client = None

    def tearDown(self):
        if self.client:
            self._await(self.client.close())
        self.loop.close()

    def _await(self, co):
        return self.loop.run_until_complete(co)

    async def _post(self, msg, asyn=False, filt=None):
        if self.client is None:
            self.client = FakeNetworkPollerGrpcClient()
        if filt:
            self.client.add_filter(filt)
        fut = await self.client.post(msg)
        if asyn:
            return None
        return await asyncio.wait_for(fut, 1)

    def post(self, msg, asyn=False, filt=None):
        return self._await(self._post(msg, asyn, filt))

    def test_network_poller_grpc_01(self):
        """Post a message.  It is delivered."""
        msg = NetworkElementUpdate()
        msg.time.FromSeconds(1612927766)
        r = self.post(msg)
        self.assertIsNotNone(r)
        self.assertEqual(len(r), 1)
        self.assertEqual(r[0], msg)

    def test_network_poller_grpc_02(self):
        """Post two messages.  They are delivered as a batch."""
        msg1 = NetworkElementUpdate()
        msg1.time.FromSeconds(1612927766)
        r = self.post(msg1, True)
        msg2 = NetworkElementUpdate()
        msg2.time.FromSeconds(1612928035)
        r = self.post(msg2)
        self.assertIsNotNone(r)
        self.assertEqual(len(r), 2)
        self.assertEqual(r[0], msg1)
        self.assertEqual(r[1], msg2)

    def test_network_poller_grpc_03(self):
        """Test maximum batch message size."""
        start_time = 1612927766
        msg = NetworkElementUpdate()
        msg.data.name = '0123456789' * 1000  # 10KB string
        n = 0
        t = start_time
        nm = 0
        while True:
            msg.time.FromSeconds(t)
            t += 1
            n += msg.ByteSize()
            nm += 1
            if n > FakeNetworkPollerGrpcClient.MAX_MESSAGE_SIZE:
                r = self.post(msg)
                break
            else:
                self.post(msg, True)
        self.assertIsNotNone(r)
        self.assertEqual(len(r), nm - 1)  # the next one will follow later
        for i, rmsg in enumerate(r):
            msg.time.FromSeconds(start_time + i)
            self.assertEqual(rmsg, msg)
        # get the last message that overflowed in a separate batch
        r = self.post(None)
        self.assertEqual(len(r), 1)
        self.assertEqual(r[0].data, msg.data)

    def test_network_poller_grpc_04(self):
        """Test oversized message is not part of the current or next batch."""
        start_time = 1612927766
        msg = NetworkElementUpdate()
        msg.data.name = '0123456789' * 1000  # 10 KB string

        n = 0
        t = start_time
        nm = 0
        r = None
        regular_msg_size = msg.ByteSize()
        while True:
            msg.time.FromSeconds(t)
            if nm == 2:
                # 3rd message is oversized
                msg.data.name = 'OOVERSIZED' * 1000 * 500

            t += 1
            n += msg.ByteSize()
            nm += 1
            if msg.ByteSize() > FakeNetworkPollerGrpcClient.MAX_MESSAGE_SIZE:
                r = self.post(msg)
                break
            else:
                self.post(msg, True)

        self.assertIsNotNone(r)
        self.assertEqual(len(r), 2)  # the batch size should stop on 2
        # back to original reg size message reference
        msg.data.name = '0123456789' * 1000  # 10 KB string
        for i, rmsg in enumerate(r):
            msg.time.FromSeconds(start_time + i)
            #compare all message in batch are regular size with proper time
            self.assertEqual(rmsg, msg)

        # post one regular size message again to see that oversized message is not part of the next batch
        msg.time.FromSeconds(t)
        r = self.post(msg, False)
        self.assertEqual(len(r), 1)
        self.assertEqual(r[0].data, msg.data)
        msg.time.FromSeconds(start_time + 4)

    def test_filter_01(self):
        """Test that a filter can be added that modifies the message."""
        def filt(msg, _):
            msg.instance = 'test'
            return msg
        msg = NetworkElementUpdate()
        msg.time.FromSeconds(1612927766)
        r = self.post(msg, filt=filt)
        self.assertIsNotNone(r)
        self.assertEqual(len(r), 1)
        msg.instance = 'test'
        self.assertEqual(r[0], msg)


    def test_filter_02(self):
        """Test that a filter can be added that suppresses the message."""
        def filt(msg, _):
            return None
        msg = NetworkElementUpdate()
        msg.time.FromSeconds(1612927766)
        self.assertRaises(asyncio.TimeoutError, self.post, msg, filt=filt)



from dataclasses import dataclass
import logging
import os
from typing import Dict, Optional
from .netmeta_config_loader import load_config
from preseem_protobuf.network_poller import network_poller_pb2

HW_SPECS_FILENAME = "pon_specs.yaml"


@dataclass
class PonClassSpec():
    pon_technology: int
    distance: Optional[int] = None
    max_split: Optional[int] = None
    min_loss: Optional[float] = None  # dBm
    max_loss: Optional[float] = None  # dBm


def load_yaml_specs(filename):
    path = os.path.join(os.path.dirname(os.path.realpath(__file__)), filename)
    try:
        cfg = load_config(path)
        if cfg is None:
            logging.warning(f"Config file {filename} returned None")
            return {}
        return cfg
    except Exception as e:
        logging.error(f"Failed to load {filename}: {e}")
        return {}


_PON_TYPE_MAP = {
    "GPON": network_poller_pb2.PonTechnology.GPON,
    "XGS-PON": network_poller_pb2.PonTechnology.XGS_PON,
    "EPON": network_poller_pb2.PonTechnology.EPON,
    "TEN-G-EPON": network_poller_pb2.PonTechnology.TEN_G_EPON,
}


def _parse_specs_to_pon_class_specs(pon_type, pon_specs):
    if not pon_specs:
        return None, None

    pon_technology = _PON_TYPE_MAP.get(pon_type)
    if pon_technology is None:
        return None, None

    pon_class_spec = PonClassSpec(
        pon_technology=pon_technology,
        distance=pon_specs.get('distance'),
        max_split=pon_specs.get('max_split'),
        min_loss=pon_specs.get('rx_power_min'),
        max_loss=pon_specs.get('rx_power_max'),
    )

    return pon_technology, pon_class_spec


def load_generic_specs():
    """
    Load the YAML input file and assign generic GPON, XGS-PON and Combo PON specs
    """
    cfg = load_yaml_specs(HW_SPECS_FILENAME)
    if not cfg:
        return {}

    class_specs = {}
    class_specs["GPON_BITRATE_DOWN"] = cfg.get("GPON_BITRATE_DOWN")
    class_specs["GPON_BITRATE_UP"] = cfg.get("GPON_BITRATE_UP")
    class_specs["XGSPON_BITRATE"] = cfg.get("XGSPON_BITRATE")
    class_specs["COMBOPON_SPLITRATIO"] = cfg.get("COMBOPON_SPLITRATIO")

    return class_specs


def load_all_hw_class_specs():
    """
    Load the YAML input file and convert the OLT/ONT PartSpecs to the PonClassSpec format.

    Use this function when loading a YAML input file with PartSpecs for OLT, ONT and PONs.
    """
    cfg = load_yaml_specs(HW_SPECS_FILENAME)

    if "PartSpecs" not in cfg:
        return {}

    class_specs = {}
    if "OLT" in cfg["PartSpecs"]:
        olt_pon_specs_dict = cfg["PartSpecs"]["OLT"]
        for olt_name, pon_types in olt_pon_specs_dict.items():
            if olt_name in class_specs:
                logging.warning(
                    f"Duplicate model name '{olt_name}' across OLT/ONT/PON in {HW_SPECS_FILENAME}"
                )
                continue
            else:
                class_specs[olt_name] = {}

            for pon_type, pon_specs in pon_types.items():
                pon_technology, pon_class_spec = _parse_specs_to_pon_class_specs(
                    pon_type, pon_specs)
                if not pon_class_spec:
                    logging.warning(
                        f"Unknown PON technology type '{pon_type}' for OLT '{olt_name}' in {HW_SPECS_FILENAME}"
                    )
                    continue

                class_specs[olt_name][pon_technology] = pon_class_spec

    if "ONT" in cfg["PartSpecs"]:
        ont_pon_specs_dict = cfg["PartSpecs"]["ONT"]
        for ont_name, pon_types in ont_pon_specs_dict.items():
            if ont_name in class_specs:
                logging.warning(
                    f"Duplicate model name '{ont_name}' across OLT/ONT/PON in {HW_SPECS_FILENAME}"
                )
                continue
            else:
                class_specs[ont_name] = {}

            for pon_type, pon_specs in pon_types.items():
                pon_technology, pon_class_spec = _parse_specs_to_pon_class_specs(
                    pon_type, pon_specs)
                if not pon_class_spec:
                    logging.warning(
                        f"Unknown PON technology type '{pon_type}' for ONT '{ont_name}' in {HW_SPECS_FILENAME}"
                    )
                    continue

                class_specs[ont_name][pon_technology] = pon_class_spec

    if "PON" in cfg["PartSpecs"]:
        pon_specs_dict = cfg["PartSpecs"]["PON"]
        for pon_name, pon_types in pon_specs_dict.items():
            if pon_name in class_specs:
                logging.warning(
                    f"Duplicate model name '{pon_name}' across OLT/ONT/PON in {HW_SPECS_FILENAME}"
                )
                continue
            else:
                class_specs[pon_name] = {}

            for pon_type, pon_specs in pon_types.items():
                pon_technology, pon_class_spec = _parse_specs_to_pon_class_specs(
                    pon_type, pon_specs)
                if not pon_class_spec:
                    logging.warning(
                        f"Unknown PON technology type '{pon_type}' for PON '{pon_name}' in {HW_SPECS_FILENAME}"
                    )
                    continue

                class_specs[pon_name][pon_technology] = pon_class_spec

    return class_specs



import asyncio
from collections import namedtuple
import logging
import time
import grpc

from preseem_grpc_model import metrics_pb2
from preseem.grpc_client import PreseemGrpcClient
from preseem.util import uint64

FlexMetric = namedtuple('FlexMetric', ('name', 'timestamp', 'labels', 'fields'))


class NetworkMetricsClient(PreseemGrpcClient):

    def __init__(self, api_key, host=None, port=None, beta=None):
        super().__init__(
            api_key, host or 'metrics{}.preseem.com'.format('-beta' if beta else ''),
            port)


class NetworkMetricsModel(object):
    _MAX_MESSAGE_SIZE = 1024 * 1024 * 2  # GRPC max message size is 4K but even getting close seems problematic?
    _MAX_LIST_LENGTH = 5000
    HOLDOFF_TIME = 10  # max seconds to wait to send a metrics batch

    def __init__(self, client):
        self._client = client
        self._flex_q = asyncio.Queue(5000)  # max size queue to limit memory
        self._flex_task = None
        self._last_q_full = None
        self._drop_count = 0

    def _flex_pb(self, flex):
        """Convert a FlexMetric object to a Flex protobuf message."""
        assert isinstance(flex, FlexMetric)
        assert isinstance(flex.labels, dict) and isinstance(flex.fields, dict)
        ts = metrics_pb2.google_dot_protobuf_dot_timestamp__pb2.Timestamp()
        ts.FromNanoseconds(int(flex.timestamp * 1000000000))
        obj = metrics_pb2.Flex(name=flex.name, timestamp=ts)
        for k, v in flex.labels.items():
            label = obj.labels.add()
            label.name = k
            label.value = v
        for f, v in flex.fields.items():
            if v is None:
                continue  # don't send null values
            tv = obj.fields[str(f)]
            tv.name = str(f)
            ty = type(v)
            if ty is bool:
                tv.type = metrics_pb2.TV.BOOL
                tv.bool_val = v
            elif ty is float:
                tv.type = metrics_pb2.TV.DOUBLE
                tv.double_val = v
            elif isinstance(v, str):
                tv.type = metrics_pb2.TV.STRING
                tv.string_val = v
            elif ty is int:
                if int.bit_length(v) > 63:
                    logging.debug(
                        "Flex: %s, labels: %s, field: %s's value: %s is bigger than max for INT64",
                        flex.name, flex.labels, f, v)
                    del obj.fields[str(f)]
                    continue
                tv.type = metrics_pb2.TV.INT64
                tv.int64_val = v
            elif ty is uint64:
                #STM-4895 - influxdb/big-query doesn't support uint64 therefore number bit length equals to or bigger than 64 are filtered
                if int.bit_length(v) >= 64:
                    logging.debug(
                        "Flex: %s, labels: %s, field: %s's value: %s is bigger than max for UINT64",
                        flex.name, flex.labels, f, v)
                    del obj.fields[str(f)]
                    continue
                if int(v) < 0:
                    logging.debug(
                        "Flex: %s, labels: %s, field: %s's value: %s cannot be negative for UINT64",
                        flex.name, flex.labels, f, v)
                    del obj.fields[str(f)]
                    continue
                tv.type = metrics_pb2.TV.UINT64
                tv.uint64_val = v
            else:
                del obj.fields[str(f)]
                raise RuntimeError('Unknown type for field {}: {}'.format(f, ty))
        return obj

    async def _service_flex_queue(self):
        """Task to consume the queue and write out flexes.  We try to batch as
           much as possible while respecting the message size and original
           batching, and without adding any delay."""
        try:
            loop = asyncio.get_event_loop()
            req = metrics_pb2.FlexPushRequest()
            size = req.ByteSize()
            buf = None  # extra flexes saved for next batch to send
            while True:
                if len(req.flexes):
                    try:
                        await self._client.push_flexes(req)
                        req = metrics_pb2.FlexPushRequest()
                        size = req.ByteSize()
                    except grpc.RpcError as rpc_error:
                        logging.warning('Error pushing flexes: %s', rpc_error)
                        await asyncio.sleep(self.HOLDOFF_TIME)
                        continue
                    except Exception as e:
                        # if it fails, we'll hold on to the failed message until
                        # it succeeds.  So we're assuming the message is ok or
                        # else we will never proceed past this.
                        logging.warning('Error pushing flexes: %s', e)
                        await asyncio.sleep(self.HOLDOFF_TIME)
                        continue
                try:
                    # Build a flex push request from the buffer and the queue
                    t = loop.time() + self.HOLDOFF_TIME
                    while True:
                        if buf:
                            pbs = buf
                            buf = None
                        else:
                            pbs = await asyncio.wait_for(self._flex_q.get(),
                                                         t - loop.time())
                            self._flex_q.task_done()
                        i = 0
                        for i, pb in enumerate(pbs):
                            if len(req.flexes) + i < self._MAX_LIST_LENGTH:
                                size += pb.ByteSize()
                                if size < self._MAX_MESSAGE_SIZE:
                                    continue
                            # send partway through the batch; buffer the rest
                            buf = pbs[i:]
                            del pbs[i:]
                            break
                        req.flexes.extend(pbs)
                        if buf or len(req.flexes) == self._MAX_LIST_LENGTH:
                            break
                except asyncio.TimeoutError:
                    pass  # we will send whatever we built on next loop iter
                except asyncio.CancelledError:
                    # try to send any in-progress message
                    try:
                        if len(req.flexes):
                            await self._client.push_flexes(req)
                    finally:
                        raise
        except asyncio.CancelledError:  # typical case
            pass
        except Exception as e:  # shouldn't happen but log if it does
            logging.info('Exiting _service_flex_queue due to %s', e)

    async def close(self):
        """close this model, cancel the background tasks."""
        if self._flex_task:
            self._flex_task.cancel()
            self._flex_task = None
        self._client = None

    async def push_flexes(self, flexes):
        """Push flex metrics.  The metrics are put on a queue and delivered
           asynchronously.  If the queue overflows they are logged out."""
        if not flexes:
            return
        assert isinstance(flexes, list)
        if not self._client:
            for flex in flexes:
                logging.warning('Metrics model closed, dropping flex: %s', flex)
            return

        if not self._flex_task:
            self._flex_task = asyncio.get_event_loop().create_task(
                self._service_flex_queue())
        try:
            self._flex_q.put_nowait([self._flex_pb(f) for f in flexes])
        except asyncio.QueueFull:  # send queue is full, drop the request
            now = time.monotonic()
            if not self._last_q_full or now - self._last_q_full > 300:
                self._last_q_full = now
                self._drop_count += len(flexes)
                logging.warning('Flex queue full, %s dropped', self._drop_count)



"""Test Source Model classes."""
import unittest
from hashlib import sha256
from uuid import uuid4

from preseem.source_model import Attachment, AccessPoint, Olt, Package, Router, Account, Service, Site
from preseem.grpc_nm_model import NetworkMetadataEntity, NetworkMetadataReference
from preseem_protobuf.model.account_pb2 import AccountStatus
from preseem_protobuf.model.common_pb2 import AccountType

class TestSourceModel(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def assertError(self, obj, field, value):
        """Check an error is properly assigned to a source object.
           We can add more specific checks as we design how this will work."""
        self.assertTrue(hasattr(obj, 'errors'))
        self.assertEqual(len(obj.errors), 1)
        err = obj.errors.pop()
        self.assertEqual(err.field, field)
        self.assertEqual(err.value, value)
        
    def test_router_id(self):
        """Test field checking of router.id"""
        with self.assertRaises(TypeError):
            Router()  # id is required
        with self.assertRaises(ValueError):
            Router(None)  # id can't be None
        with self.assertRaises(ValueError):
            Router("")  # id can't be empty string
        router = Router(123)
        self.assertEqual(router.id, "123")  # int converted to string
        router = Router(0)
        self.assertEqual(router.id, "0")  # int converted to string
        router = Router("my_id")
        with self.assertRaises(AttributeError):
            router.id = "new_id"  # id is immutable once set

    def test_router_host(self):
        """Test field checking of router.host"""
        router = Router("r1", host=None)
        self.assertEqual(router.host, None)
        router = Router("r1", host="")
        self.assertError(router, 'host', '')  # host can't be empty
        self.assertError(Router("R1", host=""), 'host', '')
        router.host = "192.0.2.1"
        self.assertEqual(router.host, "192.0.2.1")
        router.host = "192.0.2.X"
        self.assertError(router, 'host', '192.0.2.X') # invalid IP is checked
        router.host = "192.0.2.1/24"
        self.assertError(router, 'host', '192.0.2.1/24') # host can't be a subnet
        router.host = 3221225986
        self.assertEqual(router.host, "192.0.2.2")  # host is normalized
        router.host = None
        self.assertEqual(router.host, None)

    def test_router_name(self):
        router = Router("r1", name=None)
        self.assertEqual(router.name, None)
        router = Router("r1", name="")
        self.assertError(router, 'name', '')  # name can't be empty
        router = Router("r1", name='router 1')
        self.assertEqual(router.name, 'router 1')
        router.name = 123
        self.assertError(router, 'name', 123)  # name must be a string
        self.assertEqual(router.name, 'router 1')
        router.name = 'router 2'
        self.assertEqual(router.name, 'router 2')
        router.name = None
        self.assertEqual(router.name, None)

    def test_router_site(self):
        router = Router("r1", site=None)
        self.assertEqual(router.site, None)
        router = Router("r1", site="")
        self.assertError(router, 'site', '')  # site can't be empty
        router = Router("r1", site='site 1')
        self.assertEqual(router.site, 'site 1')
        router.site = 123
        self.assertError(router, 'site', 123)  # site must be a string
        self.assertEqual(router.site, 'site 1')
        router.site = 'site 2'
        self.assertEqual(router.site, 'site 2')
        router.site = None
        self.assertEqual(router.site, None)

    def test_router_url(self):
        router = Router("r1", url=None)
        self.assertEqual(router.url, None)
        router = Router("r1", url="")
        self.assertError(router, 'url', '')  # url can't be empty
        router = Router("r1", url='https://router.url')
        self.assertEqual(router.url, 'https://router.url')
        router.url = 123
        self.assertError(router, 'url', 123)  # url must be a string
        self.assertEqual(router.url, 'https://router.url')
        router.url = 'https://router.url/routers/2'
        self.assertEqual(router.url, 'https://router.url/routers/2')
        router.url = None
        self.assertEqual(router.url, None)

    def test_router_nm(self):
        """Test that the router object generates a good NetworkMetadataReference."""
        router = Router("r1")
        ref = NetworkMetadataReference('router', 'r1', {})
        self.assertEqual(router.get_nm(), [ref])
        router.host = ref.attributes['host'] = '192.0.2.1'
        self.assertEqual(router.get_nm(), [ref])
        router.name = ref.attributes['name'] = 'Router 1'
        self.assertEqual(router.get_nm(), [ref])
        router.site = ref.attributes['site'] = 'Site 1'
        self.assertEqual(router.get_nm(), [ref])
        router.community = ref.attributes['community'] = 'public'
        self.assertEqual(router.get_nm(), [ref])
        ref.attributes['system'] = 'my-system'
        self.assertEqual(router.get_nm('my-system'), [ref])
        uuid = uuid4()
        ref.attributes['uuid'] = str(uuid)
        router._uuid = uuid.bytes
        self.assertEqual(router.get_nm('my-system'), [ref])

    def test_olt(self):
        """Simple unit test for OLT, the code is shared with Router so most of the
           testing is done there."""
        olt = Olt("olt1", name='OLT 1', host='192.0.2.2', site='Site 1')
        ref = NetworkMetadataReference(type='olt', value='olt1', attributes={'host': '192.0.2.2', 'name': 'OLT 1', 'site': 'Site 1', 'system': 'test'})
        uuid = uuid4()
        ref.attributes['uuid'] = str(uuid)
        olt._uuid = uuid.bytes
        self.assertEqual(olt.get_nm('test'), [ref])

    def test_ap(self):
        """Simple unit test for AP, the code is shared with Router so most of the
           testing is done there; the main difference is the expanded set of network
           metadata used for APs."""
        ap = AccessPoint("ap1", name='AP 1', host='192.0.2.2', site='Site 1')
        ref = NetworkMetadataReference(type='ap', value='ap1', attributes={'ip_address': '192.0.2.2', 'topology1': 'AP 1', 'topology2': 'Site 1', 'system': 'test', 'topology1_snmp': 'ap1', 'ap_config': 'ap1'})
        uuid = uuid4()
        ref.attributes['uuid'] = str(uuid)
        ap._uuid = uuid.bytes
        nms = ap.get_nm('test')

        self.assertEqual(len(nms), 3)
        ap_ref = next((x for x in nms if isinstance(x, NetworkMetadataReference) and x.type == 'ap'), None)
        self.assertEqual(ap_ref, ref)
        ref = NetworkMetadataReference('session', '192.0.2.2', {'service_rate_down': 10000000, 'service_rate_up': 10000000, 'subscriber_identifier': '192.0.2.2', 'subscriber_name': 'AP 1', 'system': 'test'})
        session_ref = next((x for x in nms if isinstance(x, NetworkMetadataReference) and x.type == 'session'), None)
        self.assertEqual(session_ref, ref)
        entity = next((x for x in nms if isinstance(x, NetworkMetadataEntity)), None)
        self.assertEqual(entity.network_prefix, '192.0.2.2')
        self.assertEqual(entity.attributes, {'session': '192.0.2.2', 'snmp_session': '192.0.2.2'})

    def test_account_id(self):
        """Test field checking of account.id"""
        with self.assertRaises(TypeError):
            Account()  # id is required
        with self.assertRaises(ValueError):
            Account(None)  # id can't be None
        with self.assertRaises(ValueError):
            Account("")  # id can't be empty string
        account = Account(123)
        self.assertEqual(account.id, "123")  # int converted to string
        account = Account(0)
        self.assertEqual(account.id, "0")  # int converted to string
        account = Account("my_id")
        with self.assertRaises(AttributeError):
            account.id = "new_id"  # id is immutable once set

    def test_account_type(self):
        """Test field checking of account.type"""
        account = Account("a1", type=None)
        self.assertEqual(account.type, None)
        account = Account("a1", type="")
        self.assertError(account, 'type', '')  # type can't be a string
        account = Account("a1", type=0)
        self.assertError(account, 'type', 0)  # type can't be set to the zero enum
        account = Account("a1", type=-1)
        self.assertError(account, 'type', -1)  # type can't be set to a non-enum value
        account = Account("a1", type=AccountType.ACCOUNT_TYPE_COMMERCIAL)
        self.assertEqual(account.type, AccountType.ACCOUNT_TYPE_COMMERCIAL)
        account.type = 'RESIDENTIAL'
        self.assertError(account, 'type', 'RESIDENTIAL')  # type must be an int
        self.assertEqual(account.type, AccountType.ACCOUNT_TYPE_COMMERCIAL)
        account.type = AccountType.ACCOUNT_TYPE_RESIDENTIAL
        self.assertEqual(account.type, 2)
        account.type = None
        self.assertEqual(account.type, None)

    def test_account_status(self):
        """Test field checking of account.status"""
        account = Account("a1", status=None)
        self.assertEqual(account.status, None)
        account = Account("a1", status="")
        self.assertError(account, 'status', '')  # status can't be a string
        account = Account("a1", status=0)
        self.assertError(account, 'status', 0)  # status can't be set to the zero enum
        account = Account("a1", status=-1)
        self.assertError(account, 'status', -1)  # status can't be set to a non-enum value
        account = Account("a1", status=AccountStatus.ACCOUNT_STATUS_INACTIVE)
        self.assertEqual(account.status, AccountStatus.ACCOUNT_STATUS_INACTIVE)
        account.status = 'ACTIVE'
        self.assertError(account, 'status', 'ACTIVE')  # status must be an int
        self.assertEqual(account.status, AccountStatus.ACCOUNT_STATUS_INACTIVE)
        account.status = AccountStatus.ACCOUNT_STATUS_ACTIVE
        self.assertEqual(account.status, 3)
        account.status = None
        self.assertEqual(account.status, None)

    def test_account_name(self):
        """Test field checking of account.name"""
        account = Account("a1", name=None)
        self.assertEqual(account.name, None)
        account = Account("a1", name="")
        self.assertError(account, 'name', '')  # name can't be empty
        account = Account("a1", name='account 1')
        self.assertEqual(account.name, 'account 1')
        account.name = 123
        self.assertError(account, 'name', 123)  # name must be a string
        self.assertEqual(account.name, 'account 1')
        account.name = 'account 2'
        self.assertEqual(account.name, 'account 2')
        account.name = None
        self.assertEqual(account.name, None)

    def test_account_first_name(self):
        """Test field checking of account.first_name"""
        account = Account("a1", first_name=None)
        self.assertEqual(account.first_name, None)
        account = Account("a1", first_name="")
        self.assertError(account, 'first_name', '')  # first_name can't be empty
        account = Account("a1", first_name='fname')
        self.assertEqual(account.first_name, 'fname')
        account.first_name = 123
        self.assertError(account, 'first_name', 123)  # first_name must be a string
        self.assertEqual(account.first_name, 'fname')
        account.first_name = 'fname2'
        self.assertEqual(account.first_name, 'fname2')
        account.first_name = None
        self.assertEqual(account.first_name, None)

    def test_account_last_name(self):
        """Test field checking of account.last_name"""
        account = Account("a1", last_name=None)
        self.assertEqual(account.last_name, None)
        account = Account("a1", last_name="")
        self.assertError(account, 'last_name', '')  # last_name can't be empty
        account = Account("a1", last_name='lname')
        self.assertEqual(account.last_name, 'lname')
        account.last_name = 123
        self.assertError(account, 'last_name', 123)  # last_name must be a string
        self.assertEqual(account.last_name, 'lname')
        account.last_name = 'lname2'
        self.assertEqual(account.last_name, 'lname2')
        account.last_name = None
        self.assertEqual(account.last_name, None)

    def test_account_url(self):
        """Test field checking of account.url"""
        account = Account("a1", url=None)
        self.assertEqual(account.url, None)
        account = Account("a1", url="")  # url can't be empty string
        self.assertError(account, 'url', "")
        self.assertEqual(account.url, None)
        account = Account("a1")
        self.assertEqual(account.url, None)
        account.url = "https://my.url.com"
        self.assertEqual(account.url, "https://my.url.com")
        account = Account("s1", url="https://my.url.com")
        self.assertEqual(account.url, "https://my.url.com")
        account.url = "https://url2.net"
        self.assertEqual(account.url, "https://url2.net")
        account.url = 0
        self.assertError(account, 'url', 0)
        self.assertEqual(account.url, "https://url2.net")

    def test_account_equivalence(self):
        """Test that equivalence works for Account."""
        a1 = Account('a1', type=1, status=2, name='Full Name')
        a2 = Account('a1', type=1, status=2, name='Full Name')
        a3 = Account('a1', type=1, status=2, first_name='Full', last_name='Name')
        self.assertEqual(a1, a2)
        self.assertNotEqual(a2, a3)

    def test_service_id(self):
        """Test field checking of service.id"""
        with self.assertRaises(TypeError):
            Service()  # id is required
        with self.assertRaises(ValueError):
            Service(None)  # id can't be None
        with self.assertRaises(ValueError):
            Service("")  # id can't be empty string
        service = Service(123)
        self.assertEqual(service.id, "123")  # int converted to string
        service = Service(0)
        self.assertEqual(service.id, "0")  # int converted to string
        service = Service("my_id")
        with self.assertRaises(AttributeError):
            service.id = "new_id"  # id is immutable once set

    def test_service_account_id(self):
        """Test field checking of service.account_id"""
        service = Service("s1", account_id=None)
        self.assertEqual(service.account_id, None)
        service = Service("s1", account_id="")  # id can't be empty string
        self.assertError(service, 'account_id', "")
        self.assertEqual(service.account_id, None)
        service = Service("s1")
        self.assertEqual(service.account_id, None)
        service.account_id = "a1"
        self.assertEqual(service.account_id, "a1")
        service = Service("s1", account_id="a1")
        self.assertEqual(service.account_id, "a1")
        service.account_id = "a2"
        self.assertEqual(service.account_id, "a2")
        service.account_id = 0
        self.assertEqual(service.account_id, "0")
        service.account_id = 123
        self.assertEqual(service.account_id, "123")

    def test_service_package_id(self):
        """Test field checking of service.package_id"""
        service = Service("s1", package_id=None)
        self.assertEqual(service.package_id, None)
        service = Service("s1", package_id="")  # id can't be empty string
        self.assertError(service, 'package_id', "")
        self.assertEqual(service.package_id, None)
        service = Service("s1")
        self.assertEqual(service.package_id, None)
        service.package_id = "a1"
        self.assertEqual(service.package_id, "a1")
        service = Service("s1", package_id="a1")
        self.assertEqual(service.package_id, "a1")
        service.package_id = "a2"
        self.assertEqual(service.package_id, "a2")
        service.package_id = 0
        self.assertEqual(service.package_id, "0")
        service.package_id = 123
        self.assertEqual(service.package_id, "123")

    def test_service_download_rate(self):
        """Test field checking of service.download_rate"""
        service = Service("s1")
        self.assertEqual(service.download_rate, None)
        service = Service("s1", download_rate=None)
        self.assertEqual(service.download_rate, None)
        service = Service("s1", download_rate=0)
        self.assertEqual(service.download_rate, 0)
        service.download_rate = 1000
        self.assertEqual(service.download_rate, 1000)
        service.download_rate = "lots"
        self.assertError(service, 'download_rate', 'lots')
        self.assertEqual(service.download_rate, 1000)
        service.download_rate = 1.7
        self.assertEqual(service.download_rate, 1)

    def test_service_upload_rate(self):
        """Test field checking of service.upload_rate"""
        service = Service("s1")
        self.assertEqual(service.upload_rate, None)
        service = Service("s1", upload_rate=None)
        self.assertEqual(service.upload_rate, None)
        service = Service("s1", upload_rate=0)
        self.assertEqual(service.upload_rate, 0)
        service.upload_rate = 1000
        self.assertEqual(service.upload_rate, 1000)
        service.upload_rate = "lots"
        self.assertError(service, 'upload_rate', 'lots')
        self.assertEqual(service.upload_rate, 1000)
        service.upload_rate = 1.7
        self.assertEqual(service.upload_rate, 1)

    def test_service_subscriber_identifier(self):
        """Test field checking of service.subscriber_identifier"""
        service = Service("s1", subscriber_identifier=None)
        self.assertEqual(service.subscriber_identifier, None)
        service = Service("s1", subscriber_identifier="")  # id can't be empty string
        self.assertError(service, 'subscriber_identifier', "")
        self.assertEqual(service.subscriber_identifier, None)
        service = Service("s1")
        self.assertEqual(service.subscriber_identifier, None)
        service.subscriber_identifier = "a1"
        self.assertEqual(service.subscriber_identifier, "a1")
        service = Service("s1", subscriber_identifier="a1")
        self.assertEqual(service.subscriber_identifier, "a1")
        service.subscriber_identifier = "a2"
        self.assertEqual(service.subscriber_identifier, "a2")
        service.subscriber_identifier = 0
        self.assertEqual(service.subscriber_identifier, "0")
        service.subscriber_identifier = 123
        self.assertEqual(service.subscriber_identifier, "123")

    def test_service_url(self):
        """Test field checking of service.url"""
        service = Service("s1", url=None)
        self.assertEqual(service.url, None)
        service = Service("s1", url="")  # url can't be empty string
        self.assertError(service, 'url', "")
        self.assertEqual(service.url, None)
        service = Service("s1")
        self.assertEqual(service.url, None)
        service.url = "https://my.url.com"
        self.assertEqual(service.url, "https://my.url.com")
        service = Service("s1", url="https://my.url.com")
        self.assertEqual(service.url, "https://my.url.com")
        service.url = "https://url2.net"
        self.assertEqual(service.url, "https://url2.net")
        service.url = 0
        self.assertError(service, 'url', 0)
        self.assertEqual(service.url, "https://url2.net")

    def test_service_attachment(self):
        """Test field checking of service.attachment"""
        service = Service("s1")
        self.assertEqual(service.attachment, Attachment())
        service = Service("s1", attachment=None)
        self.assertEqual(service.attachment, None)
        with self.assertRaises(TypeError):
            service = Service("s1", attachment=1)
        with self.assertRaises(TypeError):
            service = Service("s1", attachment="attachment")
        attachment = Attachment()
        service = Service("s1", attachment=attachment)
        self.assertEqual(service.attachment, attachment)
        self.assertEqual(service.attachment.pb, attachment.pb)
        self.assertIsNot(service.attachment, attachment)  # it should make a copy

    def test_service_attachment_error_propagation(self):
        """Errors on the attachment bubble up to the service."""
        service = Service("s1")
        attachment = service.attachment
        self.assertIsNotNone(attachment)
        self.assertEqual(attachment, Attachment())
        attachment.cpe_mac_address = ""
        self.assertError(service, 'cpe_mac_address', "")
        attachment.mac_addresses.append("00:11:22:33:44:ZZ")
        self.assertError(service, 'mac_addresses', "00:11:22:33:44:ZZ")

    def test_service_equivalence(self):
        """Test that equivalence works for Service."""
        attachment1 = Attachment(cpe_mac_address='00:11:22:33:44:55')
        attachment2 = Attachment(cpe_mac_address='00:11:22:33:44:55')
        srv1 = Service('s1', account_id='a1', package_id='p1', download_rate=100000, upload_rate=20000, subscriber_identifier='a1.s1', attachment=attachment1)
        srv2 = Service('s1', account_id='a1', package_id='p1', download_rate=100000, upload_rate=20000, subscriber_identifier='a1.s1', attachment=attachment2)
        self.assertEqual(srv1, srv2)

    def test_attachment_cpe_mac_address(self):
        """Test field checking of attachment.cpe_mac_address"""
        attachment = Attachment()
        self.assertEqual(attachment, Attachment())
        attachment = Attachment(cpe_mac_address="")
        self.assertError(attachment, 'cpe_mac_address', "")
        attachment.cpe_mac_address = ""
        self.assertError(attachment, 'cpe_mac_address', "")
        attachment = Attachment(cpe_mac_address=1)  # value must be a string
        self.assertError(attachment, 'cpe_mac_address', 1)
        attachment.cpe_mac_address = 1
        self.assertError(attachment, 'cpe_mac_address', 1)
        attachment = Attachment(cpe_mac_address='zz')  # value must be a mac address
        self.assertError(attachment, 'cpe_mac_address', "zz")
        attachment.cpe_mac_address = 'zz'
        self.assertError(attachment, 'cpe_mac_address', "zz")
        attachment = Attachment(cpe_mac_address = '00:00:00:00:00:00')  # value can't be the zero mac
        self.assertError(attachment, 'cpe_mac_address', "00:00:00:00:00:00")
        attachment.cpe_mac_address = '00:00:00:00:00:00'
        self.assertError(attachment, 'cpe_mac_address', "00:00:00:00:00:00")
        attachment = Attachment()
        attachment.cpe_mac_address = '00:11:22:33:44:55'
        self.assertEqual(Attachment(cpe_mac_address='001122334455'), attachment)
        attachment.cpe_mac_address = '00:11:22:33:44:66'
        self.assertNotEqual(Attachment(cpe_mac_address='00:11:22:33:44:55'), attachment)

    def test_attachment_imsi(self):
        """Test field checking of attachment.imsi"""
        attachment = Attachment(imsi="")
        self.assertError(attachment, 'imsi', "")
        attachment.imsi = ""
        self.assertError(attachment, 'imsi', "")
        attachment = Attachment(imsi=1)
        self.assertError(attachment, 'imsi', 1)  # value must be a string
        attachment.imsi = 1
        self.assertError(attachment, 'imsi', 1)
        attachment.imsi = 310150123456789
        self.assertError(attachment, 'imsi', 310150123456789)
        attachment.imsi = '310150123456789'
        self.assertEqual(Attachment(imsi='310150123456789'), attachment)
        attachment.imsi = '515020987654321'
        self.assertNotEqual(Attachment(imsi='310150123456789'), attachment)

    def test_attachment_cpe_serial_number(self):
        """Test field checking of attachment.cpe_serial_number"""
        attachment = Attachment(cpe_serial_number="")  # value can't be empty string
        self.assertError(attachment, 'cpe_serial_number', "")
        attachment.cpe_serial_number = ""
        self.assertError(attachment, 'cpe_serial_number', "")
        attachment = Attachment(cpe_serial_number=1)  # value must be a string
        self.assertError(attachment, 'cpe_serial_number', 1)
        attachment.cpe_serial_number = 1
        self.assertError(attachment, 'cpe_serial_number', 1)
        attachment = Attachment(cpe_serial_number=310150123456789)
        self.assertError(attachment, 'cpe_serial_number', 310150123456789)
        attachment = Attachment()
        attachment.cpe_serial_number = '310150123456789'
        self.assertEqual(Attachment(cpe_serial_number='310150123456789'), attachment)
        attachment.cpe_serial_number = '515020987654321'
        self.assertNotEqual(Attachment(cpe_serial_number='310150123456789'), attachment)

    def test_attachment_cpe_name(self):
        """Test field checking of attachment.cpe_name"""
        attachment = Attachment(cpe_name="")  # value can't be empty string
        self.assertError(attachment, 'cpe_name', "")
        attachment.cpe_name = ""
        self.assertError(attachment, 'cpe_name', "")
        attachment = Attachment(cpe_name=1)
        self.assertError(attachment, 'cpe_name', 1)
        attachment.cpe_name = 1
        self.assertError(attachment, 'cpe_name', 1)
        attachment.cpe_name = 310150123456789
        self.assertError(attachment, 'cpe_name', 310150123456789)
        attachment = Attachment()
        attachment.cpe_name = '310150123456789'
        self.assertEqual(Attachment(cpe_name='310150123456789'), attachment)
        attachment.cpe_name = '515020987654321'
        self.assertNotEqual(Attachment(cpe_name='310150123456789'), attachment)

    def test_attachment_device_mac_address(self):
        """Test field checking of attachment.device_mac_address"""
        attachment = Attachment(device_mac_address="")  # value can't be empty string
        self.assertError(attachment, 'device_mac_address', "")
        attachment.device_mac_address = ""
        self.assertError(attachment, 'device_mac_address', "")
        attachment = Attachment(device_mac_address=1)  # value must be a string
        self.assertError(attachment, 'device_mac_address', 1)
        attachment.device_mac_address = 1
        self.assertError(attachment, 'device_mac_address', 1)
        attachment = Attachment(device_mac_address='zz')  # value must be a mac address
        self.assertError(attachment, 'device_mac_address', "zz")
        attachment.device_mac_address = 'zz'
        self.assertError(attachment, 'device_mac_address', "zz")
        attachment = Attachment(device_mac_address = '00:00:00:00:00:00')  # value can't be the zero mac
        self.assertError(attachment, 'device_mac_address', "00:00:00:00:00:00")
        attachment.device_mac_address = '00:00:00:00:00:00'
        self.assertError(attachment, 'device_mac_address', "00:00:00:00:00:00")
        attachment = Attachment()
        attachment.device_mac_address = '00:11:22:33:44:55'
        self.assertEqual(Attachment(device_mac_address='001122334455'), attachment)
        attachment.device_mac_address = '00:11:22:33:44:66'
        self.assertNotEqual(Attachment(device_mac_address='00:11:22:33:44:55'), attachment)

    def test_attachment_username(self):
        """Test field checking of attachment.username"""
        attachment = Attachment(username="")  # value can't be empty string
        self.assertError(attachment, 'username', "")
        attachment.username = ""
        self.assertError(attachment, 'username', "")
        attachment = Attachment(username=1)
        self.assertError(attachment, 'username', 1)
        attachment.username = 1
        self.assertError(attachment, 'username', 1)
        attachment.username = 310150123456789
        self.assertError(attachment, 'username', 310150123456789)
        attachment = Attachment()
        attachment.username = '310150123456789'
        self.assertEqual(Attachment(username='310150123456789'), attachment)
        attachment.username = '515020987654321'
        self.assertNotEqual(Attachment(username='310150123456789'), attachment)

    def test_attachment_mac_addresses(self):
        """Test field checking of attachment.mac_addresses"""
        with self.assertRaises(AttributeError):
            Attachment(mac_addresses=None)  # can't set repeated element directly
        with self.assertRaises(AttributeError):
            Attachment().mac_addresses = []
        attachment = Attachment()
        attachment.mac_addresses.append("00:11:22:33:44:ZZ")
        self.assertError(attachment, 'mac_addresses', "00:11:22:33:44:ZZ")
        self.assertFalse(attachment.mac_addresses)
        attachment.mac_addresses.append(1)
        self.assertError(attachment, 'mac_addresses', 1)
        self.assertFalse(attachment.mac_addresses)
        attachment.mac_addresses.append("")
        self.assertError(attachment, 'mac_addresses', "")
        self.assertFalse(attachment.mac_addresses)
        attachment.mac_addresses.append(None)
        self.assertError(attachment, 'mac_addresses', None)
        self.assertFalse(attachment.mac_addresses)
        attachment.mac_addresses.append('00:00:00:00:00:00')
        self.assertError(attachment, 'mac_addresses', '00:00:00:00:00:00')
        self.assertFalse(attachment.mac_addresses)
        attachment1 = Attachment()
        attachment1.mac_addresses.append('00:11:22:33:44:55')
        attachment2 = Attachment()
        attachment2.mac_addresses.extend(['00:11:22:33:44:55'])
        self.assertEqual(attachment1, attachment2)
        attachment2.mac_addresses.append('00:11:22:33:44:66')
        self.assertNotEqual(attachment1, attachment2)
        self.assertEqual(len(attachment1.mac_addresses), 1)
        self.assertEqual(len(attachment2.mac_addresses), 2)
        self.assertEqual(attachment2.mac_addresses[1], '00:11:22:33:44:66')
        del attachment2.mac_addresses[1]
        attachment1.mac_addresses[0] = '001122334455'
        self.assertEqual(attachment1, attachment2)

    def test_attachment_network_prefixes(self):
        """Test field checking of attachment.network_prefixes"""
        with self.assertRaises(AttributeError):
            Attachment(network_prefixes=None)  # can't set repeated element directly
        with self.assertRaises(AttributeError):
            Attachment().network_prefixes = []
        attachment = Attachment()
        attachment.network_prefixes.append("XYZ")
        self.assertError(attachment, 'network_prefixes', "XYZ")
        self.assertFalse(attachment.network_prefixes)
        attachment.network_prefixes.append(1.5)
        self.assertError(attachment, 'network_prefixes', 1.5)
        self.assertFalse(attachment.network_prefixes)
        attachment.network_prefixes.append("")
        self.assertError(attachment, 'network_prefixes', "")
        self.assertFalse(attachment.network_prefixes)
        attachment.network_prefixes.append(None)
        self.assertError(attachment, 'network_prefixes', None)
        self.assertFalse(attachment.network_prefixes)
        attachment.network_prefixes.append('0.0.0.0/0')
        self.assertError(attachment, 'network_prefixes', '0.0.0.0/0')
        self.assertFalse(attachment.network_prefixes)
        attachment.network_prefixes.append('::/0')
        self.assertError(attachment, 'network_prefixes', '::/0')
        self.assertFalse(attachment.network_prefixes)
        attachment1 = Attachment()
        attachment1.network_prefixes.append('192.0.2.1')
        attachment2 = Attachment()
        attachment2.network_prefixes.extend(['192.0.2.1/32'])
        self.assertEqual(attachment1, attachment2)
        attachment2.network_prefixes.append('2001:0DB8:1234:0000:0000:0000:0000:0000/48')
        self.assertNotEqual(attachment1, attachment2)
        self.assertEqual(len(attachment1.network_prefixes), 1)
        self.assertEqual(len(attachment2.network_prefixes), 2)
        self.assertEqual(attachment2.network_prefixes[1], '2001:db8:1234::/48')
        del attachment2.network_prefixes[0]
        attachment1.network_prefixes[0] = '2001:db8:1234::/48'
        self.assertEqual(attachment1, attachment2)

    def test_attachment_network_prefixes_no_dups(self):
        attachment1 = Attachment()
        attachment1.network_prefixes.append('192.0.2.1')
        self.assertEqual(len(attachment1.network_prefixes), 1)
        self.assertEqual(attachment1.network_prefixes[0], '192.0.2.1')

        attachment1.network_prefixes.append('192.0.2.1')
        self.assertEqual(len(attachment1.network_prefixes), 1)
        self.assertEqual(attachment1.network_prefixes[0], '192.0.2.1')

        attachment1 = Attachment()
        attachment1.network_prefixes.extend(['192.0.2.1/32', '192.0.2.1'])
        self.assertEqual(len(attachment1.network_prefixes), 1)
        self.assertEqual(attachment1.network_prefixes[0], '192.0.2.1')

        attachment1.network_prefixes.extend(['192.0.2.1/32', '192.0.2.1'])
        self.assertEqual(len(attachment1.network_prefixes), 1)
        self.assertEqual(attachment1.network_prefixes[0], '192.0.2.1')

    def test_attachment_mac_addresses_no_dups(self):
        mac = '00:11:22:33:44:55'
        attachment1 = Attachment()
        attachment1.mac_addresses.append(mac)
        self.assertEqual(len(attachment1.mac_addresses), 1)
        self.assertEqual(attachment1.mac_addresses[0], mac)

        attachment1.mac_addresses.append(mac)
        self.assertEqual(len(attachment1.mac_addresses), 1)
        self.assertEqual(attachment1.mac_addresses[0], mac)

        attachment1 = Attachment()
        attachment1.mac_addresses.extend([mac, '0:11:22:33:44:55'])
        self.assertEqual(len(attachment1.mac_addresses), 1)
        self.assertEqual(attachment1.mac_addresses[0], mac)

        attachment1.mac_addresses.extend([mac, '0:11:22:33:44:55'])
        self.assertEqual(len(attachment1.mac_addresses), 1)
        self.assertEqual(attachment1.mac_addresses[0], mac)

    def test_attachment_mac_addresses_order(self):
        """order of MAC addresses doesn't matter"""
        mac1 = '00:11:22:33:44:55'
        mac2 = '00:11:22:33:44:66'
        attachment1 = Attachment()
        attachment1.mac_addresses.append(mac1)
        attachment1.mac_addresses.append(mac2)

        attachment2 = Attachment()
        attachment2.mac_addresses.append(mac2)
        attachment2.mac_addresses.append(mac1)

        self.assertEqual(attachment1, attachment2)

    def test_attachment_network_prefix_order(self):
        """order of network prefixesdoesn't matter"""
        ip1 = '192.0.2.1'
        ip2 = '192.0.2.2'
        attachment1 = Attachment()
        attachment1.network_prefixes.append(ip1)
        attachment1.network_prefixes.append(ip2)

        attachment2 = Attachment()
        attachment2.network_prefixes.append(ip2)
        attachment2.network_prefixes.append(ip1)

        self.assertEqual(attachment1, attachment2)

    def test_site_id(self):
        """Test field checking of site.id"""
        with self.assertRaises(TypeError):
            Site()  # id is required
        with self.assertRaises(ValueError):
            Site(None)  # id can't be None
        with self.assertRaises(ValueError):
            Site("")  # id can't be empty string
        site = Site(123)
        self.assertEqual(site.id, "123")  # int converted to string
        site = Site(0)
        self.assertEqual(site.id, "0")  # int converted to string
        site = Site("my_id")
        with self.assertRaises(AttributeError):
            site.id = "new_id"  # id is immutable once set

    def test_site_name(self):
        """Test field checking of site.name"""
        site = Site("a1", name=None)
        self.assertEqual(site.name, None)
        site = Site("a1", name="")
        self.assertError(site, 'name', '')  # name can't be empty
        site = Site("a1", name='site 1')
        self.assertEqual(site.name, 'site 1')
        site.name = 123
        self.assertError(site, 'name', 123)  # name must be a string
        self.assertEqual(site.name, 'site 1')
        site.name = 'site 2'
        self.assertEqual(site.name, 'site 2')
        site.name = None
        self.assertEqual(site.name, None)

    def test_site_network_prefixes(self):
        """Test field checking of site.network_prefixes"""
        site = Site('s1')
        self.assertEqual(site.network_prefixes, [])
        site = Site('s1', network_prefixes=None)
        self.assertEqual(site.network_prefixes, [])
        with self.assertRaises(TypeError):
            site.network_prefixes = ''  # must be a list
        site.network_prefixes.append('192.0.2.1')
        self.assertEqual(site.network_prefixes[0], '192.0.2.1')

    def test_package_id(self):
        """Test field checking of package.id"""
        with self.assertRaises(TypeError):
            Package()  # id is required
        with self.assertRaises(ValueError):
            Package(None)  # id can't be None
        with self.assertRaises(ValueError):
            Package("")  # id can't be empty string
        pkg = Package(123)
        self.assertEqual(pkg.id, "123")  # int converted to string
        pkg = Package(0)
        self.assertEqual(pkg.id, "0")  # int converted to string
        pkg = Package("my_id")
        with self.assertRaises(AttributeError):
            pkg.id = "new_id"  # id is immutable once set

    def test_package_name(self):
        """Test field checking of package.name"""
        pkg = Package("p1", name=None)
        self.assertEqual(pkg.name, None)
        pkg = Package("p1", name="")
        self.assertError(pkg, 'name', '')  # name can't be empty
        pkg = Package("p1", name='Gold Plan')
        self.assertEqual(pkg.name, 'Gold Plan')
        pkg.name = 123
        self.assertError(pkg, 'name', 123)  # name must be a string
        self.assertEqual(pkg.name, 'Gold Plan')
        pkg.name = 'Silver Plan'
        self.assertEqual(pkg.name, 'Silver Plan')
        pkg.name = None
        self.assertEqual(pkg.name, None)

    def test_package_download_rate(self):
        """Test field checking of package.download_rate"""
        pkg = Package("p1")
        self.assertEqual(pkg.download_rate, None)
        pkg = Package("p1", download_rate=None)
        self.assertEqual(pkg.download_rate, None)
        pkg = Package("p1", download_rate=0)
        self.assertEqual(pkg.download_rate, 0)
        pkg.download_rate = 100000
        self.assertEqual(pkg.download_rate, 100000)
        pkg.download_rate = "lots"
        self.assertError(pkg, 'download_rate', 'lots')
        self.assertEqual(pkg.download_rate, 100000)
        pkg.download_rate = 1.7
        self.assertEqual(pkg.download_rate, 1)

    def test_package_upload_rate(self):
        """Test field checking of package.upload_rate"""
        pkg = Package("p1")
        self.assertEqual(pkg.upload_rate, None)
        pkg = Package("p1", upload_rate=None)
        self.assertEqual(pkg.upload_rate, None)
        pkg = Package("p1", upload_rate=0)
        self.assertEqual(pkg.upload_rate, 0)
        pkg.upload_rate = 50000
        self.assertEqual(pkg.upload_rate, 50000)
        pkg.upload_rate = "lots"
        self.assertError(pkg, 'upload_rate', 'lots')
        self.assertEqual(pkg.upload_rate, 50000)
        pkg.upload_rate = 1.7
        self.assertEqual(pkg.upload_rate, 1)

    def test_package_equivalence(self):
        """Test that equivalence works for Package."""
        p1 = Package('p1', name='Gold', download_rate=100000, upload_rate=50000)
        p2 = Package('p1', name='Gold', download_rate=100000, upload_rate=50000)
        p3 = Package('p1', name='Silver', download_rate=100000, upload_rate=50000)
        self.assertEqual(p1, p2)
        self.assertNotEqual(p1, p3)

    def test_package_nm(self):
        """Test that the package object generates a good NetworkMetadataReference."""
        pkg = Package("p1")
        ref = NetworkMetadataReference('package', 'p1', {})
        self.assertEqual(pkg.get_nm(), [ref])
        pkg.name = ref.attributes['service_name'] = 'Gold Plan'
        self.assertEqual(pkg.get_nm(), [ref])
        pkg.download_rate = ref.attributes['service_rate_down'] = 100000
        self.assertEqual(pkg.get_nm(), [ref])
        pkg.upload_rate = ref.attributes['service_rate_up'] = 50000
        self.assertEqual(pkg.get_nm(), [ref])
        ref.attributes['system'] = 'my-system'
        self.assertEqual(pkg.get_nm('my-system'), [ref])

    def test_element_poller_hash(self):
        """Test field checking of element.poller_hash"""
        elem = Olt("elem1", poller_hash=None)
        self.assertEqual(elem.poller_hash, None)
        elem = Olt("elem1", poller_hash="")
        self.assertEqual(elem.poller_hash, None)
        self.assertError(elem, 'poller_hash', "")
        elem.poller_hash = b'asdf'
        self.assertError(elem, 'poller_hash', b"asdf")
        poller_hash = sha256(b"asdf").digest()
        elem.poller_hash = poller_hash
        self.assertEqual(elem.poller_hash, poller_hash)



"""
This module provides a client class to connect to a Preseem GRPC service and
execute operations on it.  GRPC specific aspects are abstracted here, and
Python methods are provided.  By design, this module does not look at any
of the model objects, just the request/response objects and the rpc calls.

A large part of the responsibility of this module is to convert the third-party
GRPC library to the asyncio concurrency model.  This includes support for
streaming subscriptions, where the caller is blocked in a Subscribe call and
then has streaming messages asynchronously delivered into the event loop via
a callback.

This is designed to be easily replaceable by a stub client class for unit
testing.

All Preseem services are modeled here; I'm not super happy with that aspect
of the design, but for now it's left that way.
"""
import asyncio
import concurrent
import functools
from importlib import import_module
import logging
import time
import traceback

import grpc

from preseem_grpc_model import metrics_pb2, metrics_pb2_grpc
from preseem_grpc_model import network_metadata_pb2, network_metadata_pb2_grpc
from preseem_protobuf.config import company_configs_pb2, company_configs_pb2_grpc


class ConfigType():

    def __init__(self, name, prefix, value, pb, grpc):
        self.name = name  # the lowercase name of the enum
        self.value = value  # the int value of the enum
        self.prefix = prefix  # op prefix
        self.pb = pb  # pb2 import
        self.grpc = grpc  # pb2_grpc import
        self.svc = None  # service stub
        self.pb_type = None  # grpc type
        self.del_req = None  # delete request
        self.get_req = None  # get request
        self.set_req = None  # set request
        self.sub_req = None  # subscribe request


class Subscription(object):
    """Helper class to manage a GRPC subscription."""

    def __init__(self, client, op, req, cbk, handler=None):
        """
        Manage a GRPC subscription in asyncio.  Pass in the client object
        and the operation to call that will return the subscription iterator.
        New results are handled by running a callback in the eventloop.
        The "syn" parameter to the callback is set to True if this is the
        start of a new synchronize operation (after a GRPC reconnection).
        """
        self._client = client
        self._op = op
        self._req = req
        self._cbk = cbk
        self._handler = handler
        self._executor = concurrent.futures.ThreadPoolExecutor(max_workers=1)
        self._loop = asyncio.get_event_loop()
        self._iter = None

    def close(self):
        """Close the subscription.  This wakes up the blocked thread."""
        if self._iter:
            self._iter.cancel()

    def _run(self):
        """Read from the subscription, in a separate thread."""
        retry = False
        while True:
            # Setup the subscription
            syn = False
            if self._iter is None:
                try:
                    if retry:
                        logging.info('retry subscription to: %s', self._client.host)
                    else:
                        retry = True
                    self._iter = self._client._grpc_op_thread(self._op, self._req)
                except Exception as e:
                    logging.warning('Error creating subscription: %s', e)
                    time.sleep(10)
                    continue
                syn = True

            try:
                if self._handler:
                    for op in self._handler(self._iter):
                        try:
                            fut = asyncio.run_coroutine_threadsafe(
                                self._cbk(*op), self._loop)
                            result = fut.result()
                        except concurrent.futures.CancelledError:  # eventloop shutdown
                            return
                        except Exception as e:
                            print('Error calling callback: {}'.format(e))
                            traceback.print_tb(e.__traceback__)
                            time.sleep(10)

                else:
                    msg = self._iter.next()
            except grpc.RpcError as e:
                # https://grpc.io/grpc/python/grpc.html?highlight=method#grpc.StatusCode
                # Disconnect, need to restart subscription.  Restart
                # the initializing state after this.
                if e.code() == grpc.StatusCode.CANCELLED and e.details(
                ) == 'Locally cancelled by application!':
                    return  # cancelled by application
                elif e.code() == grpc.StatusCode.NOT_FOUND:
                    # STM-4023 the thing we are subscribed to has gone away.
                    logging.warning('GRPC NotFound error: %s, exit subscription',
                                    e.details())
                    break
                logging.warning('GRPC ERROR: %s: %s for host: %s', e.code(),
                                e.details(), self._client.host)
                logging.info('Reconnect to %s after 10s', self._client.host)
                time.sleep(10)  # TODO Backoff
                self._iter = None
                continue

            except Exception as e:
                print('Error reading from subscription iterator: {}'.format(e))
                print(type(e))
                time.sleep(10)  # TODO Backoff
                continue
            if not self._handler:
                try:
                    fut = asyncio.run_coroutine_threadsafe(self._cbk(msg, syn),
                                                           self._loop)
                    result = fut.result()
                except concurrent.futures.CancelledError:  # eventloop shutdown
                    return
                except Exception as e:
                    print('Error calling callback: {}'.format(e))
                    traceback.print_tb(e.__traceback__)
                    time.sleep(10)

    async def run(self):
        """Read from the subscription and block the caller on the eventloop."""
        await self._loop.run_in_executor(self._executor, self._run)


class PreseemGrpcClient(object):
    """Implement the GRPC client to the Preseem API, adapted to asyncio."""

    def __init__(self, api_key, host=None, port=None, compression=False, insecure=None):
        self.api_key = api_key
        self._executor = concurrent.futures.ThreadPoolExecutor(max_workers=10)
        self.host = host or 'nmapi.preseem.com'
        self.port = port or 443
        self.compression = compression
        self.insecure = insecure
        self._channel = None
        self._metrics_svc = None
        self._nm_svc = None
        self._nm_ref_svc = None
        self._config2_svc = None

        self._nm_ent_subscription = None
        self._nm_ref_subscription = None
        self._cfg_subscriptions = {}
        self._cfg_lock = None

        self.config_types = {}
        for ctname, ctid in company_configs_pb2.Type.items():
            if ctid < 2:
                continue  # ignore NODE configs for now
            try:
                cn = ctname.lower()
                pmod = import_module(f'preseem_protobuf.config.{cn}_pb2')
                gmod = import_module(f'preseem_protobuf.config.{cn}_pb2_grpc')
                ct = self.config_types[ctid] = ConfigType(cn, cn.replace('_', ''), ctid,
                                                          pmod, gmod)
                # assume naming conventions for the operations
                capname = ''.join([x[:1].upper() + x[1:] for x in cn.split('_')])
                ct.pb_type = getattr(pmod, f'{capname}')
                ct.get_req = getattr(pmod, f'{capname}GetRequest')
                ct.set_req = getattr(pmod, f'{capname}SetRequest')
                ct.del_req = getattr(pmod, f'{capname}DeleteRequest')
                ct.sub_req = getattr(pmod, f'{capname}SubscribeRequest')
            except ModuleNotFoundError:
                # we don't support this, that's ok
                pass

    def close_subscriptions(self):
        if self._nm_ent_subscription:
            self._nm_ent_subscription.close()
        if self._nm_ref_subscription:
            self._nm_ref_subscription.close()
        for sub in self._cfg_subscriptions.values():
            sub.close()
        self._cfg_subscriptions.clear()

    def _grpc_op_thread(self, op, *args, num_retries=5):
        for retry_num in range(num_retries + 1):
            if retry_num > 0:
                time.sleep(2**retry_num)
                logging.debug('Retrying after error result')
            try:
                return op(*args)  #, timeout=600) -- careful, not good for subscribe!
            except Exception as err:
                logging.exception('Error calling grpc service')

    def _grpc_op_collect(self, op, req, metadata, handler=None):
        """
        Run a GRPC operation and collect the results in this thread.
        If a handler function is passed in, that function will get an interator
        and can process it however it wants; whatever that function returns will be
        returned to the caller.
        """
        r = op(req, metadata=metadata)
        try:
            iterator = iter(r)  # this is to detect whether its a sequence
            if handler:
                return handler(iterator)
            return [x for x in iterator]  # this does network i/o
        except TypeError:  # not an iterable result, ok
            pass
        return r

    async def _grpc_op(self,
                       op,
                       req,
                       num_retries=5,
                       timeout=None,
                       collect_results=None,
                       metadata=None,
                       handler=None):
        loop = asyncio.get_event_loop()
        for retry_num in range(num_retries + 1):
            if retry_num > 0:
                await asyncio.sleep(2**retry_num)
                logging.debug('Retrying after error result')
            try:
                if timeout:
                    try:
                        if collect_results:
                            return await asyncio.wait_for(loop.run_in_executor(
                                self._executor, self._grpc_op_collect, op, req,
                                metadata, handler),
                                                          timeout=timeout)
                        else:
                            return await asyncio.wait_for(loop.run_in_executor(
                                self._executor,
                                functools.partial(op, req, metadata=metadata)),
                                                          timeout=timeout)
                    except asyncio.TimeoutError:
                        logging.warning('GRPC operation timed out: %s', op)
                        raise
                else:
                    if collect_results:
                        x = await loop.run_in_executor(self._executor,
                                                       self._grpc_op_collect, op, req,
                                                       metadata, handler)
                    else:
                        x = await loop.run_in_executor(
                            self._executor, functools.partial(op,
                                                              req,
                                                              metadata=metadata))
                    return x
            except grpc.RpcError as err:
                logging.debug('Error calling service: %s (%s)', err.code(),
                              err.details())
                if err.code() in (grpc.StatusCode.INTERNAL, grpc.StatusCode.NOT_FOUND,
                                  grpc.StatusCode.UNKNOWN,
                                  grpc.StatusCode.UNIMPLEMENTED,
                                  grpc.StatusCode.INVALID_ARGUMENT,
                                  grpc.StatusCode.PERMISSION_DENIED,
                                  grpc.StatusCode.RESOURCE_EXHAUSTED):
                    raise
                if retry_num == num_retries:
                    logging.warning('Error calling service: %s (%s)', err.code(),
                                    err.details())
                    raise
            except Exception as e:
                logging.warning('%s error calling service: %s', type(e).__name__, e)
                if retry_num == num_retries:
                    logging.exception('Error calling grpc service')
                    raise

    def _connect(self):
        """Connect to the server and authenticate. (blocking)"""

        class ApiKeyCallCredentials(grpc.AuthMetadataPlugin):

            def __init__(self, api_key):
                self.api_key = api_key

            def __call__(self, context, callback):
                callback((('api_key', self.api_key), ), None)

        class AuthInterceptor(
                grpc.UnaryUnaryClientInterceptor,
                grpc.UnaryStreamClientInterceptor,
                grpc.StreamUnaryClientInterceptor,
                grpc.StreamStreamClientInterceptor,
        ):
            """Universal interceptor that adds metadata to all RPC types."""

            def __init__(self, plugin):
                self.plugin = plugin

            def _augment_metadata(self, client_call_details):
                """Helper: calls your plugin and merges metadata."""
                md = []

                def callback(metadata, error):
                    if metadata:
                        md.extend(metadata)

                # Run the plugin to populate md
                self.plugin(None, callback)

                # Merge with any existing metadata
                combined = list(client_call_details.metadata or []) + md
                return client_call_details._replace(metadata=tuple(combined))

            # Unary → Unary
            def intercept_unary_unary(self, continuation, client_call_details, request):
                new_details = self._augment_metadata(client_call_details)
                return continuation(new_details, request)

            # Unary → Stream
            def intercept_unary_stream(self, continuation, client_call_details,
                                       request):
                new_details = self._augment_metadata(client_call_details)
                return continuation(new_details, request)

            # Stream → Unary
            def intercept_stream_unary(self, continuation, client_call_details,
                                       request_iterator):
                new_details = self._augment_metadata(client_call_details)
                return continuation(new_details, request_iterator)

            # Stream → Stream
            def intercept_stream_stream(self, continuation, client_call_details,
                                        request_iterator):
                new_details = self._augment_metadata(client_call_details)
                return continuation(new_details, request_iterator)

        dst = '{}:{}'.format(self.host, self.port)
        while self._nm_svc is None:
            try:
                # STM-3013, STM-3115 make grpc connection robust
                grpc_opts = [
                    ('grpc.keepalive_time_ms', 30000),  # send keepalive every 30s
                    ('grpc.keepalive_timeout_ms',
                     15000),  # keepalive ping timeout after 15s
                    ('grpc.keepalive_permit_without_calls',
                     True),  # send keepalives when no grpc calls
                    ('grpc.http2.max_pings_without_data', 0),
                    ('grpc.max_receive_message_length', 10 * 1024 * 1024),
                ]  # allow unlimited pings without data
                if self.compression:
                    grpc_opts.append(
                        ('grpc.default_compression_algorithm', grpc.Compression.Gzip))
                if self.api_key and not self.insecure:
                    call_creds = grpc.metadata_call_credentials(
                        ApiKeyCallCredentials(self.api_key))
                    try:
                        with open('/etc/pki/tls/certs/ca-bundle.crt', 'rb') as f:
                            ca_certs = f.read()
                    except:  # not fedora?  use default built-in CA bundle
                        ca_certs = None
                    ssl_creds = grpc.ssl_channel_credentials(root_certificates=ca_certs)
                    channel_creds = grpc.composite_channel_credentials(
                        ssl_creds, call_creds)
                    channel = grpc.secure_channel(dst, channel_creds, options=grpc_opts)
                else:
                    channel = grpc.insecure_channel(dst, options=grpc_opts)
                    if self.api_key:
                        channel = grpc.intercept_channel(
                            channel,
                            AuthInterceptor(ApiKeyCallCredentials(self.api_key)))
                self._channel = channel
                self._metrics_svc = metrics_pb2_grpc.NetworkMetricsStub(channel)
                self._nm_svc = network_metadata_pb2_grpc.NetworkMetadataServiceStub(
                    channel)
                self._nm_ref_svc = network_metadata_pb2_grpc.NetworkMetadataReferenceServiceStub(
                    channel)
                self._config2_svc = company_configs_pb2_grpc.CompanyConfigsServiceStub(
                    channel)
                for ct in self.config_types.values():
                    cls = [x for x in dir(ct.grpc) if x[-4:] == 'Stub']
                    if cls:
                        ct.svc = getattr(ct.grpc, cls[0])(channel)
            except:
                logging.exception('Error connecting to grpc service')
                time.sleep(30)

    async def connect(self):
        """Connect to the server and authenticate."""
        if self._cfg_lock is None:
            self._cfg_lock = asyncio.Lock()
        async with self._cfg_lock:
            await asyncio.get_event_loop().run_in_executor(self._executor,
                                                           self._connect)

    async def nm_ent_subscribe(self, cbk, company_id=None, resolve=False):
        """Block in this coroutine and call cbk as notifications arrive."""
        if not self._nm_svc:
            await self.connect()
        req = network_metadata_pb2.NetworkMetadataSubscribeRequest(
            allow_ipv6_and_shorter_ipv4_prefixes=True)
        req.do_not_resolve_references = not resolve
        if company_id:
            req.company_id = company_id
        self._nm_ent_subscription = Subscription(self, self._nm_svc.Subscribe, req, cbk)
        try:
            await self._nm_ent_subscription.run()
        except asyncio.CancelledError:
            return

    async def nm_ref_subscribe(self, cbk, company_id=None, handler=None):
        """Subscribe to Network Metadata Reference subscriptions."""
        if not self._nm_ref_svc:
            await self.connect()
        req = network_metadata_pb2.NetworkMetadataReferenceSubscribeRequest()
        if company_id:
            req.company_id = company_id
        self._nm_ref_subscription = Subscription(self,
                                                 self._nm_ref_svc.Subscribe,
                                                 req,
                                                 cbk,
                                                 handler=handler)
        try:
            await self._nm_ref_subscription.run()
        except asyncio.CancelledError:
            return

    async def nm_list(self, company_id=None):
        if not self._nm_svc:
            await self.connect()
        req = network_metadata_pb2.NetworkMetadataListRequest()
        if company_id:
            req.company_id = company_id
        return await self._grpc_op(self._nm_svc.List, req)

    async def nm_set(self, entity, company_id=None):
        if not self._nm_svc:
            await self.connect()
        req = network_metadata_pb2.NetworkMetadataSetRequest(entity=entity)
        if company_id:
            req.company_id = company_id
        return await self._grpc_op(self._nm_svc.Set, req)

    async def nm_del(self, entity, company_id=None):
        if not self._nm_svc:
            await self.connect()
        req = network_metadata_pb2.NetworkMetadataDeleteRequest(entity=entity)
        if company_id:
            req.company_id = company_id
        return await self._grpc_op(self._nm_svc.Delete, req)

    async def nm_ref_list(self, company_id=None):
        if not self._nm_ref_svc:
            await self.connect()
        req = network_metadata_pb2.NetworkMetadataReferenceListRequest()
        if company_id:
            req.company_id = company_id
        return await self._grpc_op(self._nm_ref_svc.List, req)

    async def nm_ref_set(self, ref, company_id=None):
        if not self._nm_svc:
            await self.connect()
        req = network_metadata_pb2.NetworkMetadataReferenceSetRequest(entity=ref)
        if company_id:
            req.company_id = company_id
        return await self._grpc_op(self._nm_ref_svc.Set, req)

    async def nm_ref_del(self, ref, company_id=None):
        if not self._nm_svc:
            await self.connect()
        req = network_metadata_pb2.NetworkMetadataReferenceDeleteRequest(entity=ref)
        if company_id:
            req.company_id = company_id
        return await self._grpc_op(self._nm_ref_svc.Delete, req)

    async def push_flexes(self, req):
        if not self._metrics_svc:
            await self.connect()
        return await self._grpc_op(self._metrics_svc.PushFlexes, req, timeout=300)

    async def get_config(self, config_type, config_id, company_id=None):
        """Get a config instance by type and id"""
        ct = self.config_types.get(config_type)
        if ct is None:
            raise ValueError('Unsupported config type')
        if not ct.svc:
            await self.connect()
        req = ct.get_req()
        req.company_id = company_id
        setattr(req, f'{ct.prefix}_id', config_id)
        r = await self._grpc_op(ct.svc.Get, req)
        if bool(r.success):
            return getattr(r, ct.prefix)
        else:  # probably a race where it was deleted before we could get it
            logging.warning('Failed to get %s config "%s"', ct.name, config_id)
            return None

    async def set_config(self, config_type, obj, company_id=None):
        ct = self.config_types.get(config_type)
        if ct is None:
            raise ValueError('Unsupported config type')
        if not ct.svc:
            await self.connect()
        req = ct.set_req()
        req.company_id = company_id
        getattr(req, ct.prefix).CopyFrom(obj)
        r = await self._grpc_op(ct.svc.Set, req)
        return bool(r.success)

    async def del_config(self, config_type, config_id, company_id=None):
        ct = self.config_types.get(config_type)
        if ct is None:
            raise ValueError('Unsupported config type')
        if not ct.svc:
            await self.connect()
        req = ct.del_req()
        req.company_id = company_id
        setattr(req, f'{ct.prefix}_id', config_id)
        r = await self._grpc_op(ct.svc.Delete, req)
        return bool(r.success)

    async def sub_config(self, config_type, config_id, cbk, company_id=None):
        """
        Subscribe to notifications for updates to a config instance.
        The callback takes args (company_id, config_type, message).
        """

        async def _cbk(msg, _):
            nonlocal company_id, config_type, cbk
            await cbk(company_id, config_type, msg)

        ct = self.config_types.get(config_type)
        if not ct.svc:
            await self.connect()
        req = ct.sub_req()
        req.company_id = company_id
        setattr(req, f'{ct.prefix}_id', config_id)
        sub = self._cfg_subscriptions[config_id] = Subscription(
            self, ct.svc.Subscribe, req, _cbk)
        try:
            await sub.run()
        except asyncio.CancelledError:
            sub.close()
            try:
                del self._cfg_subscriptions[config_id]
            except KeyError:
                pass  # already deleted

    async def sub_company_configs(self, cbk, company_id=None, subscribe_all=None):
        """Subscribe to company config updates."""
        if not self._config2_svc:
            await self.connect()
        req = company_configs_pb2.CompanyConfigSubscribeRequest()
        if company_id:
            req.company_id = company_id
        elif subscribe_all:
            req.subscribe_all = True
        sub = Subscription(self, self._config2_svc.Subscribe, req, cbk)
        try:
            await sub.run()
        except asyncio.CancelledError:
            sub.close()



import unittest
import importlib.util
import os
import sys
from unittest.mock import patch
import tempfile
from preseem.netmeta_config_loader import load_config, DuplicateKeyError
from mock import patch


class TestLoadConfig(unittest.TestCase):

    def setUp(self):
        self.invalid_yaml = """
        # This is a random comment
        snmp:
            community: Abc Xyz
            credentials:
                - username: user123
                  password: pass123
                  role: ap
                  username: NEWuser123
                  password: NEWpass234
                  role: NEWap
                            """

        self.valid_yaml = """
        snmp:
            community: abcdef
            credentials:
                - username: user123
                  password: pass123
                - username: user123
                  password: pass234
                - username: user234
                  password: pass567
                            """

        # Create temporary testfile for invalid config
        self.invalid_config_file = tempfile.NamedTemporaryFile(delete=False,
                                                               suffix=".yaml")
        self.invalid_config_file.write(self.invalid_yaml.encode('utf-8'))
        self.invalid_config_file_path = self.invalid_config_file.name
        self.invalid_config_file.close()

        # Create temporary testfile for valid config
        self.valid_config_file = tempfile.NamedTemporaryFile(delete=False,
                                                             suffix=".yaml")
        self.valid_config_file.write(self.valid_yaml.encode('utf-8'))
        self.valid_config_file_path = self.valid_config_file.name
        self.valid_config_file.close()

    def tearDown(self):
        # Clean up the temporary files after tests
        if os.path.exists(self.invalid_config_file.name):
            os.remove(self.invalid_config_file.name)
        if os.path.exists(self.valid_config_file.name):
            os.remove(self.valid_config_file.name)

    def test_valid_config(self):
        config = load_config(self.valid_config_file_path)
        self.assertIsInstance(config, dict)  # Valid

        # Assert all fields
        expected_config = {
            "snmp": {
                "community":
                "abcdef",
                "credentials": [
                    {
                        "username": "user123",
                        "password": "pass123"
                    },
                    {
                        "username": "user123",
                        "password": "pass234"
                    },
                    {
                        "username": "user234",
                        "password": "pass567"
                    },
                ]
            }
        }
        self.assertEqual(config, expected_config)

    @patch("preseem.netmeta_config_loader.logging")
    def test_invalid_config(self, mock_logger):
        config = load_config(self.invalid_config_file_path)

        # Check if the ambiguous YAML error message was logged
        logged_errors = [
            call[1][0] for call in mock_logger.method_calls if call[0] == "error"
        ]
        self.assertTrue(
            any("Ambiguous YAML detected" in error for error in logged_errors),
            "Expected error message not found in logged errors.")

    def test_missing_file(self):
        with self.assertRaises(RuntimeError) as context:
            load_config("random_file.yaml")
        self.assertIn("config file not found", str(context.exception))


if __name__ == "__main__":
    unittest.main()



"""SSL Probe Functionality"""
import asyncio
from collections import namedtuple
import ssl

from cryptography import x509
from cryptography.x509.oid import NameOID

# Setup a SSL context that will connect to anything
_ssl = ssl.create_default_context()
_ssl.check_hostname = False
_ssl.verify_mode = ssl.CERT_NONE
try:
    try:
        _ssl.minimum_version = ssl.TLSVersion.TLSv1  # STM-6823
    except AttributeError:
        pass  # Python 3.6
    _ssl.set_ciphers('DEFAULT@SECLEVEL=1')  # STM-4531
except ssl.SSLError:
    pass  # some platforms don't allow this, ok

SslProbeResult = namedtuple('SslProbeResult', ('error', 'org', 'cn'))


async def ssl_probe(host, port):
    """Try to connect to an SSL endpoint, and return certificate data."""
    try:
        loop = asyncio.get_running_loop()
    except AttributeError:
        # This is running python 3.6 or earlier; we don't support it.
        return SslProbeResult('UnsupportedPythonVersion', None, None)
    try:
        transport, proto = await asyncio.wait_for(loop.create_connection(
            asyncio.Protocol, host, port, ssl=_ssl, ssl_handshake_timeout=3.0),
                                                  timeout=5.0)
        sslobj = transport.get_extra_info('ssl_object')
        cert = sslobj.getpeercert(binary_form=True)
        cn = o = None
        if cert:
            cert = x509.load_der_x509_certificate(cert)
            cn = cert.subject.get_attributes_for_oid(NameOID.COMMON_NAME)
            cn = cn[0].value if cn else None
            o = cert.subject.get_attributes_for_oid(NameOID.ORGANIZATION_NAME)
            o = o[0].value if o else None
        transport.close()
        return SslProbeResult(None, o, cn)
    except ValueError:
        # some bad field in the cert, e.g. STM-9048
        return SslProbeResult('BadCert', None, None)
    except ConnectionAbortedError:
        # likely SSL handshake error
        return SslProbeResult('BadHandshake', None, None)
    except (asyncio.TimeoutError, ConnectionError, OSError) as e:
        return SslProbeResult('NoConnection', None, None)



"""Unit tests for counter32"""
from preseem import Counter32Sum
import unittest


class TestCounter32Sum(unittest.TestCase):

    def test_new_counter(self):
        """A new counter should return 0 until it is set."""
        self.assertEqual(int(Counter32Sum()), 0)

    def test_reset(self):
        """Calling reset restarts the counter at 0."""
        c = Counter32Sum()
        c.update(0)
        c.update(345)
        self.assertEqual(int(c), 345)
        c.reset()
        self.assertEqual(int(c), 0)

    def test_incr(self):
        """Test an incrementing counter."""
        c = Counter32Sum()
        c.update(10)
        self.assertEqual(int(c), 10)
        c.update(355)
        self.assertEqual(int(c), 355)
        c.update(355)
        self.assertEqual(int(c), 355)
        c.update(356)
        self.assertEqual(int(c), 356)

    def test_rollover(self):
        """Test counter rollover."""
        c = Counter32Sum()
        c.update(1000)
        c.update(2)
        self.assertEqual(int(c), 0xffffffff + 3)
        c.update(5)
        self.assertEqual(int(c), 0xffffffff + 6)

    def test_counterreset(self):
        """Test counter reset."""
        c = Counter32Sum()
        c.update(1000)
        c.update(5, True)
        self.assertEqual(int(c), 1005)
        c.update(10)
        self.assertEqual(int(c), 1010)



"""Firebase Push ID generator.  Converted to Python from
   https://gist.github.com/mikelehen/3596a30bd69384624c11
"""
import random
import time

PUSH_CHARS = '-0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ_abcdefghijklmnopqrstuvwxyz'
_last_push_time = 0
_last_rand = []


def generate_pushid():
    """Generate a Firebase Push ID.  Converted this algorithm to Python:
       https://gist.github.com/mikelehen/3596a30bd69384624c11"""
    global _last_push_time, _last_rand
    now = int(time.time() * 1000)
    dup_time = (now == _last_push_time)
    _last_push_time = now

    tsc = [0] * 8
    for i in range(7, -1, -1):
        tsc[i] = PUSH_CHARS[now % 64]
        now = int(now / 64)
    assert now == 0
    if dup_time:
        for i in range(11, -1, -1):
            if _last_rand[i] == 63:
                _last_rand[i] = 0
            else:
                _last_rand[i] += 1
                break
    else:
        _last_rand = [random.randint(0, 63) for i in range(12)]

    tsc.extend([PUSH_CHARS[x] for x in _last_rand])
    return ''.join(tsc)



"""
Helpers for IPv6 specific functionality
"""
import asyncio
import logging

from packaging import version

from preseem import NetworkMetadataReference, get_node_versions


class NetworkMetadataIpv6Settings:
    """
    IPv6 Settings Management.
    This class is responsible for determining whether IPv6 functionality
    (and also IPv4 subnet mapping functionality) is enabled.
    The logic for this is as follows.
      1. A setting for ipv6_enabled may be set, if this is set it is used.
      2. The network metadata model is checked for saved state.  If the
         saved state is set to enabled, we use it; once we switch to ipv6
         support, we never go back unless the config of (1) is set false.
         If the saved state is unset, we set it to ipv4.  We never return
         any result if we were unable to read from the metadata model; that
         is required for any result to be returned (else we raise an error).
      3. If the network metadata model was set to ipv4, we check the metrics
         to see if all nodes are running a version supporting ipv6.  If there
         are nodes and they all support ipv6, we enable ipv6 support.  This
         metrics polling is done periodically while we are still in ipv4 mode.
    """
    _version_getter = get_node_versions
    _lock = asyncio.Lock()
    _IPV6_VERSION = version.parse('1.11.1')
    _POLL_INTERVAL = 3600

    def __init__(self, model, telemetry_api, ipv6_enabled=None):
        self.model = model
        self.telemetry_api = telemetry_api
        if ipv6_enabled is not None:
            # we want checking this to be an error until it's explicitly set
            self.ipv6_enabled = ipv6_enabled
        self._ref = None
        self._task = None

    async def _get_saved_state(self, read_only=False):
        """Return state from metadata.  If not read_only, initialize it."""
        await self.model.update()
        refs = self.model.refs.get('preseem_state') or {}
        ref = refs.get('ipv6_settings')
        if ref:
            self._ref = ref
        elif not read_only:
            ref = NetworkMetadataReference('preseem_state', 'ipv6_settings',
                                           {'ipv6_enabled': False})
            await self.model.set_ref(ref)
            await asyncio.sleep(0.1)  # let the event update model.refs
            await self._get_saved_state(read_only=True)

    async def _poll(self):
        """Poll metrics for node versions"""
        while True:
            try:
                node_versions = await NetworkMetadataIpv6Settings._version_getter(
                    self.telemetry_api)
                if node_versions and min(node_versions.values()) >= self._IPV6_VERSION:
                    logging.info('IPv6 Capability Detected: %s', node_versions)
                    self._ref.attributes['ipv6_enabled'] = True
                    await self.model.set_ref(self._ref)
                    self.ipv6_enabled = True
                    break
            except asyncio.CancelledError:
                break
            except Exception as err:
                logging.info('%s error checking IPv6 capability: %s',
                             type(err).__name__, err)
            try:
                await asyncio.sleep(self._POLL_INTERVAL)
            except asyncio.CancelledError:
                break
        self._task = None

    async def close(self):
        """Release any resources used by this object."""
        if self._task is not None:
            self._task.cancel()
            await self._task
            self._task = None

    async def init(self):
        """Initialize the object."""
        async with self._lock:
            try:
                return self.ipv6_enabled
            except AttributeError:
                pass  # we need to initialize it
            ipv6_enabled = None
            await self._get_saved_state()
            assert self._ref
            ipv6_enabled = self._ref.attributes.get('ipv6_enabled')
            assert ipv6_enabled is not None
            if not ipv6_enabled and self._task is None:
                self._task = asyncio.get_event_loop().create_task(self._poll())
            self.ipv6_enabled = ipv6_enabled



import asyncio
import time
import unittest

from preseem_grpc_model import metrics_pb2
from preseem import NetworkMetricsModel, FlexMetric

class FakeMetricsClient(object):
    """Fake metrics client"""
    def __init__(self):
        self.flexes = []  # queue of requests pushed to the client
        self.error = None # error to throw on push

    async def push_flexes(self, req):
        if self.error:
            err = self.error
            self.error = None
            raise err
        self.flexes.append(req)


class TestMetrics(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.new_event_loop() # needed to initialize asyncio
        asyncio.set_event_loop(self.loop)
        self.client = FakeMetricsClient()
        self.model = NetworkMetricsModel(self.client)
        self.model.HOLDOFF_TIME = 1  # lower holdoff for faster tests

    def tearDown(self):
        self.loop.run_until_complete(self.model.close())
        self.loop.run_until_complete(self.loop.shutdown_asyncgens())
        self.loop.close()

    def _push_flexes(self, flexes):
        return self.loop.run_until_complete(self.model.push_flexes(flexes))

    def _wait_for_queue(self):
        """Wait for the flex model queue to be emptied."""
        self.loop.run_until_complete(self.model._flex_q.join())

    def _wait(self, sec):
        self.loop.run_until_complete(asyncio.sleep(sec))

    def check_req(self, pb, flexes):
        """Verify that a list of flexes is the same as a flex request."""
        req = metrics_pb2.FlexPushRequest()
        req.flexes.extend([self.model._flex_pb(x) for x in flexes])
        self.assertEqual(pb, req)

    def test_holdoff_1(self):
        """A small flex messages is sent after the 10s holdoff timer"""
        fm = FlexMetric('test', time.time(), {'a': 'b'}, {'f1': 100.0})
        self._push_flexes([fm])
        self._wait_for_queue()
        self._wait(1.1)
        self.assertEqual(len(self.client.flexes), 1)
        self.check_req(self.client.flexes[0], [fm])

    def test_holdoff_2(self):
        """Multiple messages are put in one request by the holdoff timer"""
        fm1 = FlexMetric('test1', time.time(), {'a': 'b'}, {'f1': 100.0})
        self._push_flexes([fm1])
        self._wait_for_queue()
        self._wait(0.1)
        fm2 = FlexMetric('test2', time.time(), {'a': 'c'}, {'f1': 101.0})
        self._push_flexes([fm2])
        self._wait(1.1)
        self.assertEqual(len(self.client.flexes), 1)
        self.check_req(self.client.flexes[0], [fm1, fm2])

    def test_large_push_1(self):
        """A push request with > 5000 messages is sent in two parts"""
        flexes = [FlexMetric('m', time.time(), {'a': 'b'}, {'f': float(i + 1)}) for i in range(5001)]
        self._push_flexes(flexes)
        self._wait_for_queue()
        self._wait(0.1)
        self.assertEqual(len(self.client.flexes), 1)
        self.check_req(self.client.flexes.pop(), flexes[:5000])
        self._wait(1.1)
        self.assertEqual(len(self.client.flexes), 1)
        self.check_req(self.client.flexes.pop(), flexes[5000:])

    def test_large_push_2(self):
        """A push request with 5000 messages is sent immediately"""
        flexes = [FlexMetric('m', time.time(), {'a': 'b'}, {'f': float(i + 1)}) for i in range(5000)]
        self._push_flexes(flexes)
        self._wait_for_queue()
        self._wait(0.1)
        self.assertEqual(len(self.client.flexes), 1)
        self.check_req(self.client.flexes.pop(), flexes)
        self._wait(1.1)
        self.assertEqual(len(self.client.flexes), 0)

    def test_large_push_3(self):
        """A push request with > 2M size is sent in two parts"""
        flexes = [FlexMetric('m', time.time(), {'a': 'b', 'big': 'c' * 10000}, {'f': float(i + 1)}) for i in range(209)]
        self._push_flexes(flexes)
        self._wait_for_queue()
        self._wait(0.1)
        self.assertEqual(len(self.client.flexes), 1)
        req = self.client.flexes.pop()
        self.check_req(req, flexes[:208])
        self._wait(1.1)
        self.assertEqual(len(self.client.flexes), 1)
        self.check_req(self.client.flexes.pop(), flexes[208:])

    def test_large_push_4(self):
        """A second push request that goes over the limits is handled properly."""
        flexes = [FlexMetric('m', time.time(), {'a': 'b'}, {'f': float(i + 1)}) for i in range(5001)]
        self._push_flexes(flexes[:1])
        self._push_flexes(flexes[1:])
        self._wait_for_queue()
        self._wait(0.1)
        self.assertEqual(len(self.client.flexes), 1)
        self.check_req(self.client.flexes.pop(), flexes[:5000])
        self._wait(1.1)
        self.assertEqual(len(self.client.flexes), 1)
        self.check_req(self.client.flexes.pop(), flexes[5000:])

    def test_multi_batch_1(self):
        """A large request that has to be split into multiple batches works."""
        flexes = [FlexMetric('m', time.time(), {'a': 'b'}, {'f': float(i + 1)}) for i in range(15001)]
        self._push_flexes(flexes)
        self._wait_for_queue()
        self._wait(0.1)
        self.assertEqual(len(self.client.flexes), 3)
        self.check_req(self.client.flexes.pop(0), flexes[:5000])
        self.check_req(self.client.flexes.pop(0), flexes[5000:10000])
        self.check_req(self.client.flexes.pop(0), flexes[10000:15000])
        self._wait(1.1)
        self.assertEqual(len(self.client.flexes), 1)
        self.check_req(self.client.flexes.pop(), flexes[15000:])

    def test_multi_batch_2(self):
        """Try a very large set of metrics to exercise STM-2666"""
        flexes = [FlexMetric('m', time.time(), {'a': 'b'}, {'f': float(i + 1)}) for i in range(50000)]
        self._push_flexes(flexes)
        self._wait_for_queue()
        self._wait(0.1)
        self.assertEqual(len(self.client.flexes), 10)

    def test_push_failure_1(self):
        """Make sure the push is retried if a failure occurs."""
        self.client.error = RuntimeError('TEST')
        fm = FlexMetric('test', time.time(), {'a': 'b'}, {'f1': 100.0})
        self._push_flexes([fm])
        self._wait_for_queue()
        self._wait(1.1)
        self.assertEqual(len(self.client.flexes), 0)
        self._wait(1)
        self.assertEqual(len(self.client.flexes), 1)
        self.check_req(self.client.flexes[0], [fm])



"""
Network Poller GRPC integration for sending NetworkElementUpdate messages
"""
import asyncio
import logging

import grpc

from .ringbuffer import RingBuffer
from preseem import PreseemGrpcClient
from preseem_protobuf.network_poller.network_poller_pb2 import (
    NetworkElementUpdate, SendUpdateBatchRequest, GetUpdatesRequest,
    NetworkElementDiscovery, SendDiscoveryBatchRequest, GetDiscoveryRequest)
from preseem_protobuf.network_poller.network_poller_pb2_grpc import NetworkPollerServiceStub


class NetworkPollerGrpcClient:
    """Client for posting NetworkElementUpdate messages."""
    BATCH_HOLDOFF_TIME = 10  # max seconds to wait to send a batch
    RETRY_HOLDOFF_TIME = 10  # time to wait to retry a failed post
    MAX_MESSAGE_SIZE = 1024 * 1024 * 3  # GRPC max message size is 4M
    MAX_MESSAGE_QLEN = 4000  # max number of messages we will queue
    POST_TIMEOUT = 30  # timeout post request after waiting 30s

    def __init__(self,
                 api_key=None,
                 host=None,
                 port=None,
                 compression=False,
                 insecure=None):
        if host:
            self._client = PreseemGrpcClient(api_key=api_key,
                                             host=host,
                                             port=port,
                                             compression=compression,
                                             insecure=insecure)
        else:
            self._client = None  # for testing
        self.host = host
        self._svc = None
        self._msg_q_neu = None  #asyncio.Queue(5000)
        self._task_neu = None
        self._filters = []
        self.drop_count_neu = 0
        self._msg_q_ned = None
        self._task_ned = None
        self.drop_count_ned = 0
        self._ringbuffer = RingBuffer("preseem-network-poller-neu")

    async def close(self):
        """Close this object and cancel the client task."""
        if self._task_neu:
            self._task_neu.cancel()
            try:
                await self._task_neu
            except asyncio.CancelledError:
                pass
            self._task_neu = None
        if self._task_ned:
            self._task_ned.cancel()
            try:
                await self._task_ned
            except asyncio.CancelledError:
                pass
            self._task_ned = None
        if self._client:
            self._client = None
        if self._ringbuffer:
            self._ringbuffer.close()

    async def get_updates(self, company_uuid=None):
        if not self._client:
            return
        if not self._svc:
            await self._client.connect()
            self._svc = NetworkPollerServiceStub(self._client._channel)
        req = GetUpdatesRequest()
        if company_uuid:
            req.company_uuid = company_uuid
        return await self._client._grpc_op(self._svc.GetUpdates,
                                           req,
                                           collect_results=True)

    async def get_discovery(self, company_uuid=None):
        if not self._client:
            return
        if not self._svc:
            await self._client.connect()
            self._svc = NetworkPollerServiceStub(self._client._channel)
        req = GetDiscoveryRequest()
        if company_uuid:
            req.company_uuid = company_uuid
        return await self._client._grpc_op(self._svc.GetDiscovery,
                                           req,
                                           collect_results=True)

    async def _post_neu(self, messages):
        """Post a list of NetworkElementUpdate messages"""
        if not self._client:
            return
        if not self._svc:
            await self._client.connect()
            self._svc = NetworkPollerServiceStub(self._client._channel)
        req = SendUpdateBatchRequest()
        req.data.extend(messages)
        return await self._client._grpc_op(self._svc.SendUpdateBatch,
                                           req,
                                           timeout=self.POST_TIMEOUT)

    async def _post_ned(self, messages):
        """Post a list of NetworkElementDiscovery messages"""
        if not self._client:
            return
        if not self._svc:
            await self._client.connect()
            self._svc = NetworkPollerServiceStub(self._client._channel)
        req = SendDiscoveryBatchRequest()
        req.data.extend(messages)
        return await self._client._grpc_op(self._svc.SendDiscoveryBatch,
                                           req,
                                           timeout=self.POST_TIMEOUT)

    async def _service_queue(self, msg_q, post_fn, type):
        """Service the message queue"""
        n = 0
        try:
            loop = asyncio.get_event_loop()
            messages = []
            next_message = None
            size = 0
            while True:
                if messages:
                    try:
                        await post_fn(messages)
                        messages.clear()
                        size = 0
                        if next_message:
                            messages.append(next_message)
                            size = next_message.ByteSize()
                            next_message = None
                    except asyncio.CancelledError:
                        raise
                    except grpc.RpcError as err:
                        if type is NetworkElementUpdate:
                            logging.warning(
                                'Error [%s] posting NetworkElementUpdate to %s: %s.  Queue size: %s.',
                                err.code(), self.host, err.details(),
                                self._msg_q_neu.qsize())
                        elif type is NetworkElementDiscovery:
                            logging.warning(
                                'Error [%s] posting NetworkElementDiscovery to %s: %s.  Queue size: %s.',
                                err.code(), self.host, err.details(),
                                self._msg_q_ned.qsize())
                        if err.code() == grpc.StatusCode.RESOURCE_EXHAUSTED:
                            # Shouldn't happen, but likely too large message
                            messages.clear()
                        elif err.code(
                        ) == grpc.StatusCode.INTERNAL and 'Exception deserializing response!' in err.details(
                        ):
                            # HS-61 although we're not sure why we get this error,
                            # it seems to have been delivered, and so we should clear
                            # the queued messages to avoid processing duplicates.
                            messages.clear()
                        await asyncio.sleep(self.RETRY_HOLDOFF_TIME)
                        continue
                    except Exception as err:
                        # if it fails, we'll retry until it succeeds.  So we're
                        # assuming the message is ok or we will never proceed.
                        logging.warning('Error posting %s: %s', type.__name__, err)
                        await asyncio.sleep(self.RETRY_HOLDOFF_TIME)
                        continue
                try:
                    # Deque messages to build a batch to post
                    t = loop.time() + self.BATCH_HOLDOFF_TIME
                    while True:
                        if messages:
                            # build the batch until the holdoff time
                            msg = await asyncio.wait_for(msg_q.get(), t - loop.time())
                        else:
                            msg = await msg_q.get()
                        msg_q.task_done()
                        msg_size = msg.ByteSize()
                        # PPA-1378 first we need to check if current message is not exceeding the maximum message size, if it is, break without setting it to a next_message
                        if msg_size > self.MAX_MESSAGE_SIZE:
                            next_message = None
                            logging.warning(
                                'AP: %s, NetworkElementUpdate message with size: %s, is exceeding maximum allowed size: %s',
                                self.host, msg_size, self.MAX_MESSAGE_SIZE)
                            break
                        # break in case current message is less than maximum allowed but overall messages exceeding that. Setting current message be a next_message to form a new batch
                        elif size + msg_size > self.MAX_MESSAGE_SIZE:
                            next_message = msg
                            break
                        else:
                            size += msg_size
                            messages.append(msg)
                except asyncio.TimeoutError:
                    pass  # we will send whatever we built on next loop iter
                except asyncio.CancelledError:
                    # try to send any in-progress message
                    try:
                        if messages:
                            await post_fn(messages)
                            if next_message:
                                await post_fn([next_message])
                    finally:
                        raise
        except asyncio.CancelledError:
            pass

    def add_filter(self, f):
        """Add a filter to handle messages as they are posted.  Take (msg). return msg or None."""
        self._filters.append(f)

    async def post(self, msg, id_key=None):
        """Push a NetworkElementUpdate / NetworkElementDiscovery message.  Optionally pass in an id_key
           which will be passed along to any message filters."""
        assert isinstance(msg, (NetworkElementUpdate, NetworkElementDiscovery))
        type = None
        if isinstance(msg, NetworkElementUpdate):
            type = NetworkElementUpdate
            msg_copy = NetworkElementUpdate()
        elif isinstance(msg, NetworkElementDiscovery):
            type = NetworkElementDiscovery
            msg_copy = NetworkElementDiscovery()
        if not self._task_neu:
            self._msg_q_neu = asyncio.Queue(self.MAX_MESSAGE_QLEN)
            self._task_neu = asyncio.get_event_loop().create_task(
                self._service_queue(self._msg_q_neu, self._post_neu,
                                    NetworkElementUpdate))
        if not self._task_ned:
            self._msg_q_ned = asyncio.Queue(self.MAX_MESSAGE_QLEN)
            self._task_ned = asyncio.get_event_loop().create_task(
                self._service_queue(self._msg_q_ned, self._post_ned,
                                    NetworkElementDiscovery))
        # Make a copy to make sure the caller can't change it after this
        msg_copy.CopyFrom(msg)
        try:
            if type is NetworkElementUpdate:
                for f in self._filters:  # Apply filter only on NetworkElementUpdate
                    msg_copy = f(msg_copy, id_key)
                    if msg_copy is None:  # filter out this message
                        return
                if self._ringbuffer.exists():
                    try:
                        await self._ringbuffer.write(msg_copy.SerializeToString())
                    except TimeoutError:
                        # Semaphore timeout: this probably means the semaphore was
                        # left held by some process exiting while it was held.  Unlikely
                        # but possible.  In this case we will simply log/drop, since
                        # we need to send our message.
                        logging.warning("Semaphore timeout teeing update message")
                self._msg_q_neu.put_nowait(msg_copy)
            elif type is NetworkElementDiscovery:
                self._msg_q_ned.put_nowait(msg_copy)
        except asyncio.QueueFull:  # send queue is full, drop the request
            if type is NetworkElementUpdate:
                self.drop_count_neu += 1
                if str(self.drop_count_neu)[0] != str(self.drop_count_neu - 1)[0]:
                    # log when we're dropping, but don't spam the log
                    logging.warning('NetworkElementUpdate queue full, %s dropped',
                                    self.drop_count_neu)
            elif type is NetworkElementDiscovery:
                self.drop_count_ned += 1
                if str(self.drop_count_ned)[0] != str(self.drop_count_ned - 1)[0]:
                    # log when we're dropping, but don't spam the log
                    logging.warning('NetworkElementDiscovery queue full, %s dropped',
                                    self.drop_count_ned)


class FakeNetworkPollerGrpcClient(NetworkPollerGrpcClient):
    """Fake network poller client.  Will print the message if verbose.
       Post returns a future that will be set when the message is sent."""
    BATCH_HOLDOFF_TIME = 0.1  # short delay for testing

    def __init__(self, verbose=0, compression=False):
        super().__init__(host=None, compression=compression)
        self.verbose = verbose
        self.fut = None

    async def _post_neu(self, messages):
        """Post a list of NetworkElementUpdate messages"""
        # TODO pretty-print messages if verbose > 0
        if self.fut:
            self.fut.set_result(messages.copy())
            self.fut = None

    async def _post_ned(self, messages):
        """Post a list of NetworkElementDiscovery messages"""
        # TODO pretty-print messages if verbose > 0
        if self.fut:
            self.fut.set_result(messages.copy())
            self.fut = None

    async def post(self, msg, id_key=None):
        if msg:
            await super().post(msg, id_key)
        else:  # we can pass a null message to just wait for a second result
            pass
        fut = self.fut or asyncio.get_event_loop().create_future()
        self.fut = fut
        return fut



"""
Thin wrapper for requesting TCP latency metrics
"""
from importlib import import_module
import os
import pkgutil

from preseem import PreseemGrpcClient
from ._apiutil import load_api, op, op_sync
import preseem_protobuf.metrics


class PreseemMetrics:
    """API to Preseem Metrics."""

    def __init__(self, client: PreseemGrpcClient) -> None:
        """Initialize with a GRPC client."""
        self._client = client
        self.cache_enabled = False
        self._objs = {}
        self._types = {}

    def _load(self, wrapper):
        pkgpath = os.path.dirname(preseem_protobuf.metrics.__file__)
        for x in pkgutil.iter_modules([pkgpath]):
            if x.name.endswith('_pb2_grpc'):
                name = x.name[:-9]  # trim "_pb2_grpc" from the name
                # Load the python modules for this model object type
                pmod = import_module(f'preseem_protobuf.metrics.{name}_pb2')
                gmod = import_module(f'preseem_protobuf.metrics.{name}_pb2_grpc')
                load_api(self, wrapper, pmod, gmod)

    async def init(self):
        """Dynamically load TCP latency protobuf modules."""
        await self._client.connect()
        self._load(op)

    def load(self):
        """Dynamically load TCP latency protobuf modules. (non-asyncio)"""
        self._client._connect()
        self._load(op_sync)



"""Unit tests for counter32"""
from preseem import counter32
import unittest

class TestCounter32(unittest.TestCase):
    def test_new_counter(self):
        """A new counter should return 0 until it is set."""
        self.assertEqual(int(counter32()), 0)

    def test_reset(self):
        """Calling reset restarts the counter at 0."""
        c = counter32()
        c.update(0)
        c.update(345)
        self.assertEqual(int(c), 345)
        c.reset()
        self.assertEqual(int(c), 0)

    def test_incr(self):
        """Test an incrementing counter."""
        c = counter32()
        c.update(10)
        self.assertEqual(int(c), 0)
        c.update(355)
        self.assertEqual(int(c), 345)
        c.update(355)
        self.assertEqual(int(c), 345)
        c.update(356)
        self.assertEqual(int(c), 346)

    def test_rollover(self):
        """Test counter rollover."""
        c = counter32()
        c.update(0)
        c.update(0xfffffffe)
        self.assertEqual(int(c), 0xfffffffe)
        c.update(0xffffffff)
        self.assertEqual(int(c), 0xffffffff)
        c.update(200)
        self.assertEqual(int(c), 0xffffffff + 200)



import asyncio
from importlib import import_module
import json
import unittest

import preseem.grpc_config_model
from preseem_protobuf.config import company_configs_pb2, powercode_pb2
from preseem import ConfigModel, PreseemGrpcClient
from google.protobuf.json_format import MessageToJson

class ConfigType():
    def __init__(self, name, value, pb):
        self.name = name    # the lowercase name of the enum
        self.value = value  # the int value of the enum
        self.pb = pb        # pb2 import
        self.svc = None     # service stub
        self.pb_type = None # grpc type

def _make(nt, **kwargs):
    """Helper to make a new namedtuple object"""
    return nt(**{x: kwargs.get(x) for x in nt._fields})

class FakeConfigClient(PreseemGrpcClient):
    """Fake config client"""
    def __init__(self):
        super().__init__('TEST')
        self.pbs = {}
        self.company_id = 'TestCompany'
        self._cbk = None
        self._new_cbk = None
        self._obj_cbk = {}
        self._pc_cbk = None

    def close_subscriptions(self):
        self._cbk = None

    async def get_config(self, config_type, config_id, company_id=None):
        return self.pbs[(company_id, config_type, config_id)]

    async def del_config(self, config_type, config_id, company_id=None):
        company_id = company_id or self.company_id  # have to have a company id
        try:
            del self.pbs[(company_id, config_type, config_id)]
            if self._cbk:
                ev = company_configs_pb2.CompanyConfigEvent(company_id=company_id, config_id=config_id, event_type=company_configs_pb2.CompanyConfigEvent.DELETE, config_type=config_type)
                try:
                    await self._cbk(ev, False)
                except Exception as err:
                    print("BAD", err)
            return True
        except KeyError:
            pass

    async def sub_config(self, config_type, config_id, cbk, company_id=None):
        company_id = company_id or self.company_id
        self._obj_cbk[(company_id, config_type, config_id)] = cbk
        pb = self.pbs[(company_id, config_type, config_id)]
        if pb:
            await cbk(company_id, config_type, pb)

    async def set_config(self, config_type, obj, company_id=None):
        company_id = company_id or self.company_id  # have to have a company id
        ct = self.config_types[config_type]
        cur = self.pbs.get((company_id, config_type, obj.id))
        self.pbs[(company_id, config_type, obj.id)] = obj
        cbk = self._obj_cbk.get((company_id, config_type, obj.id))
        if cbk:
            await cbk(company_id, config_type, obj)
        if self._cbk:
            op = company_configs_pb2.CompanyConfigEvent.UPDATE if cur else company_configs_pb2.CompanyConfigEvent.SET
            ev = company_configs_pb2.CompanyConfigEvent(company_id=company_id, config_id=obj.id, event_type=op, config_type=config_type)
            await self._cbk(ev, False)
        return True

    async def list_config_instances(self, company_id=None):
        r = instances_pb2.InstancesListResponse(success=True)
        r.items.extend([instances_pb2.TypeID(type=x[0], id=y) for x in self.pbs.items() for y in x[1].keys()])
        return r

    async def sub_config_instances(self, cbk, types, company_id=None):
        self._cbk = cbk

    async def sub_company_configs(self, cbk, company_id=None, subscribe_all=None):
        async def task(self):
            try:
                while True:
                    await asyncio.sleep(1)
            except asyncio.CancelledError:
                print("CANCELME")
                self._cbk = None
                print("CANCEL!", self._cbk)
        self._cbk = cbk
        for company_id, config_type, config_id in self.pbs:
            msg = company_configs_pb2.CompanyConfigEvent(company_id=company_id, config_id=config_id, event_type=company_configs_pb2.CompanyConfigEvent.SET, config_type=config_type)
            await self._cbk(msg)
        # XXX doesn't work yet, this shoudl be done in bg after returning.
        done_msg = company_configs_pb2.CompanyConfigEvent(event_type=company_configs_pb2.CompanyConfigEvent.LOADED)
        await self._cbk(done_msg)
        return asyncio.get_event_loop().run_task()


class TestConfig(unittest.TestCase):

    def setUp(self):
        self.loop = asyncio.new_event_loop() # needed to initialize asyncio
        asyncio.set_event_loop(self.loop)
        self.client = FakeConfigClient()
        self.model = ConfigModel(self.client)

    def tearDown(self):
        self.model.close()
        self.loop.close()

    def _delete(self, cfg):
        return self.loop.run_until_complete(self.model.delete(cfg))

    def _list(self):
        return self.loop.run_until_complete(self.model.list())

    def _listen(self, cbk=None):
        return self.loop.run_until_complete(self.model.listen(cbk))

    def _set(self, cfg):
        return self.loop.run_until_complete(self.model.set(cfg))

    def _make(self, ty, **kwargs):
        """Helper to make a new config object"""
        obj = self.model.get_default('powercode')
        return obj._replace(**kwargs)

    def test_list_1(self):
        """Test basic list operations"""
        self.assertEqual(self._list(), [])
        cfg = self._make('powercode', id='test')
        self._set(cfg)
        self.assertEqual(self._list(), [cfg])

    def test_del_1(self):
        """Test deletion."""
        cfg = self._make('powercode', id='test')
        self._set(cfg)
        r = self._list()
        self.assertEqual(len(r), 1)
        r = r[0]
        self.assertEqual(r.id, cfg.id)
        self._delete(cfg)
        r = self._list()
        self.assertEqual(len(r), 0)

    def test_set_1(self):
        """Test a basic Set operation"""
        cfg = self._make('powercode', id='test')
        self._set(cfg)
        r = self._list()
        self.assertEqual(len(r), 1)
        r = r[0]
        self.assertEqual(r.id, cfg.id)

    def test_set_2(self):
        """An ID gets created if not passed in."""
        self._set(self._make('powercode'))
        r = self._list()
        self.assertEqual(len(r), 1)
        self.assertTrue(r[0].id)

    def test_set_powercode(self):
        """Test that Powercode attributes are set and obtained back via get."""
        cfg = self._make('powercode', id='test', enabled=True,
                    url='https://192.0.2.1', api_key='--apikey--',
                    block_non_active_customers=True, report_usage=True,
                    service_rate_multiplier=1.07, 
                    extra={'map_session': True, 'cpe_categories': ['Other']})
        self._set(cfg)
        r = self._list()
        self.assertEqual(len(r), 1)
        r = r[0]
        self.assertEqual(r, cfg)

    def test_subscribe_1(self):
        """The internal objs dict is kept in sync."""
        self._listen()
        cfg = self._make('powercode', id='test')
        self._set(cfg)
        self.assertEqual(len(self.model.objs), 1)
        cfg = self._make('powercode', id='test2', enabled=True,
                    url='https://192.0.2.1', api_key='--apikey--',
                    block_non_active_customers=True, report_usage=True,
                    service_rate_multiplier=1.07,
                    extra={'map_session': True, 'cpe_categories': ['Other']})
        self._set(cfg)
        self.assertEqual(len(self.model.objs), 2)
        self.assertEqual(self.model.objs.get('test2'), cfg)
        self._delete(cfg)
        self.assertEqual(len(self.model.objs), 1)
        self.assertEqual(self.model.objs.get('test2'), None)

    def test_subscribe_2(self):
        """Callbacks are sent properly."""
        async def cbk(old, new):
            nonlocal cbk_old, cbk_new
            cbk_old = old
            cbk_new = new
        self._listen(cbk)
        cfg = self._make('powercode', id='test', enabled=True,
                    url='https://192.0.2.1', api_key='--apikey--',
                    block_non_active_customers=True, report_usage=True,
                    service_rate_multiplier=1.07,
                    extra={'map_session': True, 'cpe_categories': ['Other']})
        cbk_old = cbk_new = None
        self._set(cfg)
        self.assertEqual(cbk_old, None)
        self.assertEqual(cbk_new, cfg)
        cbk_old = cbk_new = None
        newcfg = cfg._replace(enabled=False)
        self._set(newcfg)
        self.assertEqual(cbk_old, cfg)
        self.assertEqual(cbk_new, newcfg)
        cbk_old = cbk_new = None
        self._set(newcfg)
        self.assertEqual(cbk_old, None)
        self.assertEqual(cbk_new, None)
        self._delete(newcfg)
        self.assertEqual(cbk_old, newcfg)
        self.assertEqual(cbk_new, None)

    def test_ap_csvfile_1(self):
        """Basic checks for the ap-csvfile config API."""
        # Get the default value and check it
        data = self.model.get_default('ap-csvfile')
        self.assertIsNotNone(data)
        self.assertIsNotNone(getattr(data, 'url', None))
        self.assertIsNotNone(getattr(data, 'google_id', None))
        self.assertIsNotNone(getattr(data, 'delimiter', None))
        self.assertIsNotNone(getattr(data, 'header', None))
        self.assertIsNotNone(getattr(data, 'poll_interval', None))
        self.assertIsNotNone(getattr(data, 'fields', None))
        fields = data.fields
        self.assertIsInstance(data.fields, dict)
        self.assertIsNotNone(fields.get('id'))
        self.assertIsNotNone(fields.get('sector'))
        self.assertIsNotNone(fields.get('tower'))
        self.assertIsNotNone(fields.get('ip_address'))

        # Set an object
        data = data._replace(url='https://preseem.com')
        data.fields['sector'] = 'ssid'
        self._set(data)

        # Verify it
        cfgs = self._list()
        self.assertIsNotNone(cfgs)
        self.assertEqual(len(cfgs), 1)
        self.assertEqual(cfgs[0], data)

    def test_customer_csvfile_map(self):
        """Test handling of the map in the customer-csvfile protobuf."""
        data = self.model.get_default('customer-csvfile')
        data = data._replace(account_status_config={"OFFLINE": {"kbps_down": 1, "kbps_up": 1, "account_status": "DELINQUENT"}})
        self._set(data)
        cfgs = self._list()
        self.assertEqual(len(cfgs), 1)
        self.maxDiff = 10000
        obj = cfgs[0]._asdict()
        obj.pop('id')  # don't compare this, it changes from run to run 
        self.assertEqual(obj, {'enabled': False, 'extra': None, 'url': '', 'google_id': '', 'delimiter': '', 'fields': {'id': '', 'secondary_id': '', 'account_id': '', 'name': '', 'cpe_mac': '', 'dev_mac': '', 'cpe_ip_address': '', 'ip_address': '', 'kbps_down': '', 'kbps_up': '', 'mbps_down': '', 'mbps_up': '', 'package_name': '', 'sector': '', 'imsi': '', 'ip_address_2': '', 'serial_number': '', 'account_status': ''}, 'header': [], 'poll_interval': 0, 'system': '', 'service_rate_multiplier': 0.0, 'account_status_config': {'OFFLINE': {'account_status': 'DELINQUENT', 'kbps_down': 1, 'kbps_up': 1}}})



from preseem.grpc_client import PreseemGrpcClient

from preseem.grpc_nm_model import NetworkMetadataClient, PreseemNetworkMetadataModel
from preseem.grpc_nm_model import NetworkMetadataEntity, NetworkMetadataReference, NetworkMetadataIndex, NetworkMetadataLowercaseIndex
from preseem.grpc_nm_model import Reference
from preseem.fake_nm_model import FakeNetworkMetadataModel
from preseem.preseem_model import PreseemModel
from preseem.fake_preseem_model import FakePreseemModel

from preseem.preseem_metrics import PreseemMetrics

from preseem.grpc_metrics_model import FlexMetric, NetworkMetricsClient, NetworkMetricsModel
from preseem.fake_metrics_model import FakeNetworkMetricsModel

from preseem.grpc_config_model import ConfigClient, ConfigModel

from preseem.http_client import HttpClient, work_around_aiohttp_3535
from preseem.mikrotik import MikrotikApiClient
from preseem.fake_mikrotik import FakeMikrotikApiClient
from preseem.ssh_client import SshClient

from preseem.util import counter32, Counter32Sum, get_node_versions, http_probe, normalize_mac, parse_ip_value, PreseemApiKey, uint64, TelemetryApiClient, ping, set_ipv6_settings, run_tool_script, mikrotik_api_probe, ssh_probe

from preseem.ipv6 import NetworkMetadataIpv6Settings

from preseem.network_poller_grpc import NetworkPollerGrpcClient, FakeNetworkPollerGrpcClient

from preseem.load_balancer_manager_grpc import LoadBalancerManagerClient

from preseem.ssl_probe import ssl_probe

from preseem.interval_stats import UpdateIntervalManager
__version__ = "2.83.6"



"""
Tests for sentry functionality.
"""
import unittest
from preseem.sentry import SentryClient, SentryLog
from time import monotonic


class MockSentryClient(SentryClient):
    """A mock client to log errors or messages to Sentry."""

    def __init__(self, sentry_key, server_name: str):
        self.logs = {}


class TestSentryClient(unittest.TestCase):
    """Tests for Sentry Client."""

    def mock_add_log_side_effect(logs, log_type, frequency, description, tags={}):
        """Mock SentryClient._add_log"""
        log = SentryLog(log_type, str(tags), frequency, str(description))
        log_time = int(monotonic())
        last_log_time = logs.get(log)
        if not last_log_time or log_time - last_log_time >= frequency:
            if frequency > 0:
                logs[log] = log_time

    @unittest.mock.patch('preseem.sentry.SentryClient._add_log',
                         side_effect=mock_add_log_side_effect)
    def test_get_brmacs(self, mock_add_log):
        """Test logging error and message has monotonic log time."""
        sentry_client = MockSentryClient("Test Key", "Test Server")
        last_test_case = 0
        start_time = int(monotonic())

        test_log1 = SentryLog("err", "{}", 10, '1')
        test_log2 = SentryLog("msg", "{}", 0, '2')

        while (True):
            now = int(monotonic()) - start_time
            if now < 5 and last_test_case < 1:
                assert (sentry_client.logs.get(test_log1) is None)
                assert (sentry_client.logs.get(test_log2) is None)
                mock_add_log(sentry_client.logs, "err", 10, "1")
                mock_add_log(sentry_client.logs, "msg", 0, "2")
                assert (sentry_client.logs.get(test_log1) is not None)
                assert (sentry_client.logs.get(test_log2) is None)
                assert (sentry_client.logs.get(test_log1) >= now)
                last_test_case = 1
            elif 5 <= now < 10 and last_test_case < 2:
                assert (sentry_client.logs.get(test_log1) is not None)
                assert (sentry_client.logs.get(test_log2) is None)
                log_time_before = sentry_client.logs[test_log1]
                mock_add_log(sentry_client.logs, "err", 10, "1")
                mock_add_log(sentry_client.logs, "msg", 0, "2")
                assert (sentry_client.logs.get(test_log1) is not None)
                assert (sentry_client.logs.get(test_log2) is None)
                assert (sentry_client.logs[test_log1] == log_time_before)
                last_test_case = 2
            elif 10 <= now < 15 and last_test_case < 3:
                assert (sentry_client.logs.get(test_log1) is not None)
                log_time_before = sentry_client.logs[test_log1]
                mock_add_log(sentry_client.logs, "err", 10, "1")
                assert (sentry_client.logs.get(test_log1) is not None)
                assert (sentry_client.logs.get(test_log1) >= log_time_before)
                assert (sentry_client.logs.get(test_log1) >= now)
                last_test_case = 3
            elif now >= 15:
                break



"""
Mikrotik API Client.
https://wiki.mikrotik.com/wiki/Manual:API
"""
import asyncio
from binascii import hexlify, unhexlify
from collections import namedtuple
from datetime import timedelta
from hashlib import md5
import itertools
from keyword import iskeyword
import logging
import ssl
import struct

UNIT_FIELDS = {
    'w': 'weeks',
    'd': 'days',
    'h': 'hours',
    'm': 'minutes',
    's': 'seconds',
    'ms': 'milliseconds'
}


def parse_api_timedelta(s):
    """Parse the time delta format used in Mikrotik API, e.g. '2w1d2h21m2s120ms'.
       This should never raise an error; it returns None if the string is bad."""
    if not isinstance(s, str) or not s:
        return None
    parsed = {field: 0 for field in UNIT_FIELDS.values()}
    i = 0
    value = ''

    try:
        while i < len(s):
            if s[i].isdigit():
                value += s[i]
                i += 1
                continue

            if not value:
                return None

            if s[i:i + 2] == 'ms':  # special case for 2 character string
                unit = 'ms'
                i += 2
            else:
                unit = s[i]
                i += 1

            parsed[UNIT_FIELDS[unit]] += int(value)
            value = ''

        if value:
            return None
        return timedelta(**parsed)
    except Exception:
        return None


class MikrotikApiError(Exception):
    pass


class MikrotikApiClient:
    """Query the Mikrotik API"""
    _ssl_context = None

    def __init__(self,
                 host,
                 username,
                 password,
                 enable_ssl=False,
                 port=None,
                 timeout=30,
                 ignore_permissions=False,
                 ignore_internal_errors=False,
                 record=False):
        self.host = host
        self.port = port or (8729 if enable_ssl else 8728)
        self._user = username
        self._pass = password
        self._timeout = timeout
        self._ignore_internal_errors = ignore_internal_errors
        self._ignore_permissions = ignore_permissions
        self._logged_in = False
        self._reader = None
        self._writer = None
        self._ssl = None
        self.record = [] if record else None

        if enable_ssl:
            if not MikrotikApiClient._ssl_context:
                # PPA-1362 Change the way sslctx is configured
                sslctx = MikrotikApiClient._ssl_context = ssl.create_default_context()
                sslctx.check_hostname = False
                sslctx.verify_mode = ssl.CERT_NONE
                sslctx.set_ciphers('ALL:@SECLEVEL=0')
            self._ssl = MikrotikApiClient._ssl_context

        # map to filter special chars in api keys
        self.trans_map = str.maketrans('-', '_', './')
        self.error = None  # this will be set if we hit an error we can continue from

    async def _get_results_data(self,
                                cmd,
                                include_disabled=False,
                                props=None,
                                query=None):
        """Run a command and return a list of namedtuple results."""
        if not self._logged_in:
            await self.login()
        results = []
        try:
            sentence = [cmd]
            if props:  # comma-separated list of property names to return
                sentence.append(f'=.proplist={props}')
            if query:  # query words
                sentence.extend(query)
            res = await self.talk(sentence)
        except MikrotikApiError as err:
            if str(err) == 'no such command prefix':
                return []
            raise
        for r in res:
            if r[0] == '!empty':
                continue
            if r[0] == '!done':
                break
            elif not r or r[0] != '!re':
                print('unexpected result for {}: {}'.format(self.host, r))
                raise RuntimeError('unexpected result')
            row = {x[0][1:]: x[1] for x in r[1].items()}
            if include_disabled:
                results.append(row)
            else:
                if row.get('disabled') != 'true':
                    results.append(row)
        if results:
            keys = list(sorted(set(itertools.chain.from_iterable(results))))
            # keys cannot include special chars '/' , '.' or started with digit
            fields = [
                f"x_{str(x).translate(self.trans_map)}"
                if str(x[0]).isdigit() else str(x).translate(self.trans_map)
                for x in keys
            ]
            fields = [x.upper() if iskeyword(x) else x for x in fields]
            nt = namedtuple('Row', fields)
        return [nt(*[x.get(k) for k in keys]) for x in results]

    async def _get_results(self,
                           cmd,
                           include_disabled=False,
                           props=None,
                           query=None,
                           num_retries=1):
        err = None
        for retry_num in range(num_retries + 1):
            if retry_num > 0:
                await asyncio.sleep(2**retry_num)
            try:
                return await asyncio.wait_for(self._get_results_data(
                    cmd, include_disabled, props, query),
                                              timeout=self._timeout)
            except asyncio.CancelledError:
                logging.info('Cancelled "%s" operation for %s', cmd, self.host)
                await self.close()
                raise
            except asyncio.TimeoutError:
                logging.info('Operation "%s" to host %s timed out', cmd, self.host)
                await self.close()
                raise
            except Exception as e:
                if isinstance(e, MikrotikApiError):
                    if 'not enough permissions' in str(e):
                        self.error = e
                        if self._ignore_permissions:
                            logging.info(
                                'Not enough permissions for "%s" on %s, returning empty result.',
                                cmd, self.host)
                            return []  # just return empty results
                        raise  # no need to retry this
                    elif 'contact MikroTik support' in str(e):
                        self.error = e
                        if self._ignore_internal_errors:
                            logging.info(
                                '%s running command "%s" on %s, returning empty result.',
                                e, cmd, self.host)
                            return []  # just return empty results
                        raise  # no need to retry this
                logging.info('Error calling "%s" on host %s: %s (%s)', cmd, self.host,
                             type(e).__name__, e)
                err = e
                await self.close()
        logging.info('Mikrotik get "%s" to %s retried out', cmd, self.host)
        raise err

    async def connect(self):
        self._reader, self._writer = await asyncio.open_connection(self.host,
                                                                   self.port,
                                                                   ssl=self._ssl)

    async def close(self):
        if self._writer:
            try:
                self._writer.close()
            except Exception:
                pass
            self._writer = self._reader = None
            self._logged_in = False

    def encode_word(self, w):
        """Encode a word to write to the server."""
        if isinstance(w, str):
            w = w.encode('iso-8859-1')
        l = len(w)
        if l < 0x80:
            buf = struct.pack('>B', l)
        elif l < 0x4000:
            buf = struct.pack('>I', l | 0x8000)[2:]
        elif l < 0x200000:
            buf = struct.pack('>I', l | 0xc00000)[1:]
        elif l < 0x10000000:
            buf = struct.pack('>I', l | 0xe0000000)
        else:
            buf = struct.pack('>BI', 0xff, l)
        return buf + w

    async def write(self, words):
        """Write a list of "words" to the server."""
        buf = b''
        for w in words:
            buf += self.encode_word(w)
        self._writer.write(buf + b'\0')
        return len(words)

    async def login(self):
        """Login to the Mikrotik API."""
        if self._logged_in:
            return
        r = await self.talk(
            ["/login", f'=name={self._user}', f'=password={self._pass}'])
        if r:
            repl, attrs = r[0]
            if attrs:  # old style challenge:
                chal = unhexlify(attrs['=ret'])
                md = md5()
                md.update(b'\x00')
                md.update(self._pass.encode())
                md.update(chal)
                r = await self.talk([
                    "/login", f'=name={self._user}',
                    "=response=00" + hexlify(md.digest()).decode('ascii')
                ])
        self._logged_in = True

    async def talk(self, words):
        """Send words to the server and get a result back."""
        if not self._reader:
            await self.connect()
        if await self.write(words) == 0:
            return
        r = []
        err = None
        record = None if self.record is None else [words, []]
        if record:
            self.record.append(record)
        while True:
            i = await self.read()
            if record:
                record[1].append(i)
            if len(i) == 0:
                continue
            reply = i[0]
            attrs = {}
            for w in i[1:]:
                j = w.find('=', 1)
                if (j == -1):
                    attrs[w] = ''
                else:
                    attrs[w[:j]] = w[j + 1:]
            r.append((reply, attrs))
            if reply == '!done':
                if err:
                    raise err
                return r
            elif reply == '!trap':  # wait for done to return
                err = MikrotikApiError(attrs.get('=message'))
            elif reply == '!fatal':
                reason = list(attrs.keys())[0]
                if reason == "too many commands before login":
                    # PPA-923 we want to let the caller continue in this case.
                    await self.close()
                    return await self.talk(words)
                raise MikrotikApiError(reason)

    async def read(self):
        """Read a result from the API."""
        r = []
        while 1:
            w = await self.read_word()
            if w == '':
                return r
            r.append(w)

    async def read_word(self):
        """Read a string from the server."""
        # length
        buf = await self._reader.read(1)
        if not buf:  # disconnected
            await self.close()
            raise MikrotikApiError('Connection closed by remote end')
        byte_value = buf[0]
        if (byte_value & 0x80) != 0x80:
            buf = bytes([byte_value & ~0x80])
            i = 0
        elif (byte_value & 0xC0) != 0xC0:
            buf = bytes([byte_value & ~0xC0])
            i = 1
        elif (byte_value & 0xE0) != 0xE0:
            buf = bytes([byte_value & ~0xE0])
            i = 2
        elif (byte_value & 0xF0) != 0xF0:
            buf = bytes([byte_value & ~0xF0])
            i = 3
        else:
            buf = b''
            i = 4
        if i:
            buf += await self._reader.read(i)
        buf = b'\x00' * (4 - len(buf)) + buf
        length = struct.unpack('>I', buf)[0]
        # value
        ret = ''
        while len(ret) < length:
            s = await self._reader.read(length - len(ret))
            if s == b'':
                raise RuntimeError("connection closed by remote end")
            ret += s.decode('iso-8859-1')
        return ret

    async def get(self, cmd, include_disabled=False, props=None, query=None):
        """Run a command and get the results as a list of Row namedtuples."""
        return await self._get_results(cmd, include_disabled, props, query)



"""A Sentry Client Tool."""
import sentry_sdk
from collections import namedtuple
from time import monotonic
from enum import Enum

SentryLog = namedtuple("SentryLog", ("type", "tags", "frequency", "description"))


class LogType(Enum):
    ERROR = "err"
    MESSAGE = "msg"


class SentryClient:
    """A client to log errors or messages to Sentry.
    
    Attribute:
        logs: A map that records the time of last sending of SentryLog to Sentry.
    """

    def __init__(self, sentry_key, server_name: str):
        """Initializes the instance and initialize sentry_sdk based on server_name.

        Args:
            sentry_key: The Sentry DSN of the sentry project.
            server_name: A string representing the server name.
        """
        sentry_sdk.init(dsn=sentry_key,
                        traces_sample_rate=1.0,
                        attach_stacktrace=True,
                        include_local_variables=True,
                        server_name=server_name)
        self.logs: dict = {}

    def _add_log(self, log_type: LogType, frequency: int, description, tags: dict = {}):
        """Insert an error or a message to Sentry.

        Insert a SentryLog to Sentry when first time encountering or reached
        its frequency.

        Args:
            log_type:
                The type of SentryLog, either 'err' or 'msg'.
            frequency:
                The frequency (time interval) of SentryLog being sent to Sentry for
                a second time from last sending.
            description:
                The actual error or message to be sent to Sentry.
            tags:
                A dictionary of tag to tag value associated to the log.
        """
        assert isinstance(frequency, int)
        log = SentryLog(log_type, str(tags), frequency, str(description))
        log_time = int(monotonic())
        last_log_time = self.logs.get(log)
        if not last_log_time or log_time - last_log_time >= frequency:
            with sentry_sdk.push_scope() as scope:
                for tag, tag_value in tags.items():
                    scope.set_tag(tag, tag_value)
                if log_type == LogType.ERROR:
                    sentry_sdk.capture_exception(description)
                elif log_type == LogType.MESSAGE:
                    sentry_sdk.capture_message(description)
            if frequency > 0:
                self.logs[log] = log_time

    def error(self, description, frequency: int = 0, tags: dict = {}):
        """Insert an error to Sentry.

        See _add_log for more details.
        """
        assert isinstance(frequency, int)
        self._add_log(LogType.ERROR, frequency, description, tags)

    def warning(self, description: str, frequency: int = 0, tags: dict = {}):
        """Insert a warning message to Sentry.

        See _add_log for more details.
        """
        assert isinstance(frequency, int)
        self._add_log(LogType.MESSAGE, frequency, description, tags)



"""
Fake model used for testing and verification.
"""
import asyncio
import logging
from .grpc_metrics_model import FlexMetric


def _print_flex_batch(flexes):
    """default handler function"""
    from tabulate import tabulate
    print(f'{len(flexes)} Flex Metrics')
    if flexes:
        print(tabulate(flexes, flexes[0]._fields))


class FakeNetworkMetricsModel(object):
    """Fake metrics client.  You can pass in a callback that will be called
       with each metric batch that would be output to the metrics server by
       the real model.  By default it just prints them out."""

    def __init__(self, cbk=None):
        self._flex_q = asyncio.Queue(5000)  # same size queue as real impl
        self._flex_task = None
        self.cbk = cbk or _print_flex_batch
        self.drop_count = 0

    async def close(self):
        if self._flex_task:
            self._flex_task.cancel()
            self._flex_task = None

    async def _service_flex_queue(self):
        """Service the flex queue"""
        n = 0
        try:
            while True:
                flexes = await self._flex_q.get()
                n += 1
                try:
                    flexes.extend(self._flex_q.get_nowait())
                    n += 1
                except asyncio.QueueEmpty:
                    pass
                try:
                    if asyncio.iscoroutinefunction(self.cbk):
                        await self.cbk(flexes)
                    else:
                        self.cbk(flexes)
                except Exception as err:
                    logging.error('Error pushing flexes: %s', err)
                finally:
                    for _ in range(n):
                        self._flex_q.task_done()
                    n = 0
        except asyncio.CancelledError:  # typical case
            pass
        except Exception as e:
            logging.info('Exiting _service_flex_queue due to %s', e)

    async def push_flexes(self, flexes):
        """Push flex metrics."""
        if not flexes:
            return
        assert isinstance(flexes, list)
        if not self._flex_task:
            self._flex_task = asyncio.get_event_loop().create_task(
                self._service_flex_queue())
        try:
            self._flex_q.put_nowait(flexes)
        except asyncio.QueueFull:  # send queue is full, drop the request
            self.drop_count += len(flexes)



"""
Shared memory implementation of a ringbuffer.

This is designed for multiple writers and a single reader.  A semaphore is used to
lock the data strocture so that only one process can read or write to it at once.
If the ring buffer is full, the writer will overwrite the oldest messages to make room
for the new message.

This is designed so that the ring buffer lives outside of any particular process,
meaning either the reader, writer or both can be stopped or restarted without losing
queued messages.  This means the shared memory is not cleaned up by anyone; you can
manually clean it up like this:
 rm /dev/shm/{name}.
However that only works properly in Python 3.13+.  In lower Python versions, exiting
a process will delete the shared memory.

Semaphore Timeout Protection:
By default, semaphore acquisition uses a 5-second timeout to protect against orphaned
locks (when a process crashes while holding the semaphore). If a timeout occurs, a
TimeoutError is raised with instructions for manual recovery. You can disable timeouts
by passing lock_timeout=None to the constructor.
"""
import posix_ipc
import asyncio
import struct
from concurrent.futures import ThreadPoolExecutor
import time

try:
    from multiprocessing import shared_memory
    SHARED_MEMORY_AVAILABLE = True
except ImportError:
    # Python 3.7 doesn't have shared_memory in multiprocessing
    shared_memory = None
    SHARED_MEMORY_AVAILABLE = False


class RingBuffer:
    META_SIZE = 12  # Metadata: head (4 bytes), tail (4 bytes), capacity (4 bytes)
    EXISTS_CACHE_TIME = 60  # cache exists result

    def __init__(self, name, capacity=None, create=False, lock_timeout=5.0):
        self.name = name
        self.capacity = capacity
        self.create = create
        self.lock_timeout = lock_timeout  # Timeout in seconds for lock acquisition (None = block forever)
        self.shm = None
        self.lock = None
        self.executor = ThreadPoolExecutor() if SHARED_MEMORY_AVAILABLE else None
        self._exists = None  # cached result
        self._check_time = 0

    def exists(self):
        """Check if the ringbuffer shared memory exists."""
        if not SHARED_MEMORY_AVAILABLE:
            return False
        now = time.monotonic()
        if self._exists is not None and now < self._check_time:
            return self._exists
        try:
            try:
                shm = shared_memory.SharedMemory(name=self.name, track=False)
            except TypeError:
                shm = shared_memory.SharedMemory(name=self.name)
            shm.close()
            self._exists = True
        except FileNotFoundError:
            self._exists = False
            self.close()
        self._check_time = now + self.EXISTS_CACHE_TIME
        return self._exists

    def _create(self):
        """Create shared memory and semaphore."""
        size = self.META_SIZE + self.capacity  # Total shared memory size
        try:
            self.shm = shared_memory.SharedMemory(name=self.name,
                                                  create=self.create,
                                                  size=size,
                                                  track=False)
        except TypeError:  # can't control tracking in python < 3.13
            self.shm = shared_memory.SharedMemory(name=self.name,
                                                  create=self.create,
                                                  size=size)
        struct.pack_into('iii', self.shm.buf, 0, 0, 0,
                         self.capacity)  # Initialize head, tail, capacity
        self.lock = posix_ipc.Semaphore(self.name, posix_ipc.O_CREAT, initial_value=1)

    def _attach(self):
        """Create or attach shared memory"""
        if self.create:
            try:
                self._create()
                return
            except FileExistsError:
                pass  # already created, attach instead

        # Attach instead of creating
        try:
            self.shm = shared_memory.SharedMemory(name=self.name, track=False)
        except TypeError:
            self.shm = shared_memory.SharedMemory(name=self.name)
        except FileNotFoundError:
            self._exists = False
            self.close()
            return
        self.lock = posix_ipc.Semaphore(self.name)

        # Initialize capacity
        if not self.capacity:
            self._acquire_lock()
            try:
                _, _, self.capacity = self._get_metadata()
            finally:
                self.lock.release()

    def _acquire_lock(self):
        """Acquire the lock with timeout. Raises TimeoutError if lock cannot be acquired."""
        if self.lock_timeout is None:
            # Block forever (original behavior)
            self.lock.acquire()
        else:
            try:
                self.lock.acquire(self.lock_timeout)
            except posix_ipc.BusyError:
                raise TimeoutError(
                    f"Could not acquire lock '{self.name}' within {self.lock_timeout}s. "
                    "The semaphore may be orphaned (previous process crashed while holding it). "
                    "Consider manually cleaning up: rm /dev/shm/sem.{self.name}")

    def close(self):
        """Close the shared memory and shutdown executor."""
        if self.shm:
            self.shm.close()
        if self.executor:
            self.executor.shutdown(wait=False)
        self.shm = self.lock = None
        self.executor = None

    def unlink(self):
        """Unlink shared memory and semaphore."""
        try:
            if self.shm:
                self.shm.unlink()
        except FileNotFoundError:
            pass  # Already unlinked
        try:
            if self.lock:
                self.lock.unlink()
        except posix_ipc.ExistentialError:
            pass  # Already unlinked

    def _get_metadata(self):
        """Retrieve head, tail, and capacity (assumes lock is held by caller)."""
        return struct.unpack_from('iii', self.shm.buf, 0)

    def _set_metadata(self, head, tail):
        """Update head and tail pointers (assumes lock is held by caller)."""
        struct.pack_into('ii', self.shm.buf, 0, head, tail)

    def _two_phase_write(self, start: int, buf: bytes) -> None:
        """Write `buf` into the ring buffer in up to two phases if it crosses the boundary."""
        ring_start = self.META_SIZE
        ring_end = self.META_SIZE + self.capacity  # One past last valid byte
        length = len(buf)

        # Phase 1: write up to the end of the buffer
        space_to_end = ring_end - start
        if length <= space_to_end:
            # Fits in one contiguous slice
            self.shm.buf[start:start + length] = buf
        else:
            # Partially fill the end
            chunk1_size = space_to_end
            self.shm.buf[start:ring_end] = buf[:chunk1_size]

            # Phase 2: wrap to the front
            chunk2_size = length - chunk1_size
            wrap_start = ring_start  # typically the start offset of ring data
            self.shm.buf[wrap_start:wrap_start + chunk2_size] = buf[chunk1_size:]

    def _two_phase_read(self, start: int, length: int) -> bytes:
        """Read `length` bytes from the ring buffer in up to two phases if it crosses the boundary."""
        ring_start = self.META_SIZE
        ring_end = self.META_SIZE + self.capacity
        data_out = bytearray(length)

        space_to_end = ring_end - start
        if length <= space_to_end:
            # Single chunk read
            data_out[:] = self.shm.buf[start:start + length]
        else:
            # Phase 1: read until the end
            chunk1_size = space_to_end
            data_out[:chunk1_size] = self.shm.buf[start:ring_end]

            # Phase 2: wrap to the front
            chunk2_size = length - chunk1_size
            wrap_start = ring_start
            data_out[chunk1_size:] = self.shm.buf[wrap_start:wrap_start + chunk2_size]

        return bytes(data_out)

    def _write(self, data: bytes):
        """Blocking: Write a serialized message to the buffer, discarding old messages if needed."""
        if self.lock is None:
            self._attach()
        self._acquire_lock()
        try:
            head, tail, capacity = self._get_metadata()
            message_size = len(data)
            total_size = message_size + 4  # 4 bytes for the length prefix

            if total_size > capacity:
                raise ValueError("Message too large for buffer")

            # Keep discarding old messages until there's enough space
            next_tail = (tail + total_size) % capacity
            while (next_tail - head) % capacity < total_size and head != tail:
                # Read old message size at head
                prefix_start = self.META_SIZE + head
                old_size_bytes = self._two_phase_read(prefix_start, 4)
                old_message_size = struct.unpack('>I', old_size_bytes)[0]

                head = (head + old_message_size + 4) % capacity
                # Recompute next_tail after discarding
                _, tail_, _ = self._get_metadata()
                tail = tail_
                next_tail = (tail + total_size) % capacity

            # Now write the combined prefix + data in two-phase
            full_msg = struct.pack('>I', message_size) + data
            start_offset = self.META_SIZE + tail
            self._two_phase_write(start_offset, full_msg)

            # Update metadata
            self._set_metadata(head, next_tail)
        finally:
            self.lock.release()

    def _read(self):
        """Blocking: Read a serialized message from the buffer, using two-phase read if needed."""
        if self.lock is None:
            self._attach()
        self._acquire_lock()
        try:
            head, tail, capacity = self._get_metadata()
            if head == tail:
                return None  # Buffer empty

            # First read the 4-byte prefix
            prefix_start = self.META_SIZE + head
            prefix_bytes = self._two_phase_read(prefix_start, 4)
            message_size = struct.unpack('>I', prefix_bytes)[0]

            # Validate message size to prevent reading corrupted data
            if message_size == 0 or message_size > capacity:
                raise ValueError(
                    f"Corrupted ring buffer: invalid message_size={message_size}, capacity={capacity}"
                )

            # Then read the message itself
            message_start = (head + 4) % capacity
            data = self._two_phase_read(self.META_SIZE + message_start, message_size)

            # Advance head
            next_head = (head + message_size + 4) % capacity
            self._set_metadata(next_head, tail)

            return data
        finally:
            self.lock.release()

    async def get_metadata(self):
        """Async wrapper for retrieving metadata."""
        loop = asyncio.get_running_loop()

        def _get():
            self._acquire_lock()
            try:
                return self._get_metadata()
            finally:
                self.lock.release()

        return await loop.run_in_executor(self.executor, _get)

    async def write(self, data: bytes):
        """Async wrapper for writing data."""
        loop = asyncio.get_running_loop()
        try:
            await loop.run_in_executor(self.executor, self._write, data)
        except FileNotFoundError:
            self._exists = False
            self.close()

    async def read(self):
        """Async wrapper for reading data."""
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(self.executor, self._read)
